from .core import IRule
from .group_mixins.basic import create_default_logwriter
from .event_logger import tostrs_func_call


def print_method(rule: IRule):
    info = rule._get_info()  # pyright: ignore [reportPrivateUsage]
    raw_rule = info.rule_store.rules[rule.raw_rule_id]
    sl = []
    method = raw_rule.method
    tostrs_func_call(sl, method, method.args, method.kwargs)

    a = create_default_logwriter("debug")
    a.debug(*sl)



from __future__ import annotations
from multiprocessing.context import SpawnContext
import sys
import traceback
from multiprocessing import get_context
from threading import Thread, Condition

from collections import defaultdict
from typing import (
    Dict,
    Mapping,
    Set,
    Tuple,
    Callable,
    List,
    Union,
    Sequence,
    TypeVar,
)

from . import events
from .abc import IRule, IEvent
from .make import process_rule, Result, MakeSummary


def _collect_rules(
    id2rule: Sequence[IRule], seed_ids: Sequence[int]
) -> Tuple[List[int], Dict[int, Set[int]]]:
    """collect rules on which seed rules depend"""
    ids: Set[int] = set()
    b2a: Dict[int, Set[int]] = defaultdict(set)  # dict<ID, ID> before to after

    def find_deps(i: int):
        if i in ids:
            return

        ids.add(i)

        for dep in id2rule[i].deps:
            find_deps(dep)
            b2a[dep].add(i)

    for i in seed_ids:
        find_deps(i)

    return list(ids), b2a


_T_Rule = TypeVar("_T_Rule", bound=IRule)


def make_mp_spawn(
    id2rule: Sequence[_T_Rule],
    ids: Sequence[int],
    dry_run: bool,
    keep_going: bool,
    callback: Callable[[IEvent[_T_Rule]], None],
    njobs: int,
):
    """
    callback must be thread safe
    """
    if len(ids) == 0:
        return MakeSummary(total=0, update=0, skip=0, fail=0, discard=0)

    assert njobs >= 2

    # Use the process starting method 'spawn' regardless of OS
    ctx = get_context("spawn")

    # Not very confident but this useless Pool seems necessary for later
    # use of Pool in threads to work reliably.
    with ctx.Pool(1):
        pass

    # Gather relevant rules
    main_ids = ids
    ids, b2a = _collect_rules(id2rule, main_ids)

    dep_cnt = {i: len(id2rule[i].deps) for i in ids}

    # Check inter-process transferability
    rules = [id2rule[i] for i in ids]
    sendable = _test_interproc_portabability([r.method for r in rules], ctx)
    _log_sendable_stats(sendable)
    sendable = {ids[j]: sendable[j] for j in range(len(ids))}

    # state vars
    updated_ids: Set[int] = set()  # rules processed and not skipped

    nskips = 0  # for stats report
    nfails = 0  # for stats report

    job_q: List[int] = []  # FIFO: visit nodes in depth-first order

    nidles = njobs  # #idle slots

    stop = False

    cv = Condition()  # for the above state vars

    # Add rules with no dependencies to the job queue
    for i in ids:
        if dep_cnt[i] == 0:
            job_q.append(i)

    def get_job() -> Union[int, None]:
        nonlocal nidles

        with cv:
            while True:
                if stop:
                    return None
                elif len(job_q) != 0:
                    nidles -= 1
                    assert nidles >= 0
                    return job_q.pop()
                elif nidles == njobs:
                    return None

                cv.wait()

    def set_result(i: int, res: Union[Result, None]):
        nonlocal stop, nidles, nskips, nfails

        with cv:
            nidles += 1
            assert nidles <= njobs

            if res is None:  # fatal error
                nfails += 1
                stop = True
            elif res == Result.Fail:
                nfails += 1
                if not keep_going:
                    stop = True
            else:
                if res == Result.Update:
                    updated_ids.add(i)
                else:
                    nskips += 1

                for nxt in b2a[i]:
                    dep_cnt[nxt] -= 1
                    assert dep_cnt[nxt] >= 0

                    if dep_cnt[nxt] == 0:
                        job_q.append(nxt)

            cv.notify_all()

    args = (
        ctx,
        get_job,
        set_result,
        id2rule,
        updated_ids,
        main_ids,
        sendable,
        dry_run,
        callback,
    )

    threads = [Thread(target=worker, args=args) for _ in range(njobs)]

    for t in threads:
        t.start()

    try:
        for t in threads:
            t.join()
    except Exception:
        pass
    finally:
        with cv:
            stop = True
            cv.notify_all()

    return MakeSummary(
        total=len(ids),
        update=len(updated_ids),
        skip=nskips,
        fail=nfails,
        discard=len(ids) - (len(updated_ids) + nskips + nfails),
    )


def worker(
    ctx: SpawnContext,
    get_job: Callable[[], Union[int, None]],
    set_result: Callable[[int, Union[Result, None]], None],
    id2rule: List[_T_Rule],
    updated_ids: Set[int],
    main_ids: Set[int],
    sendable: Mapping[int, bool],
    dry_run: bool,
    callback: Callable[[IEvent[_T_Rule]], None],
):
    with ctx.Pool(1) as pool:
        while True:
            i = get_job()

            if i is None:
                return

            rule = id2rule[i]

            res = None

            try:
                par_updated = any(dep in updated_ids for dep in rule.deps)
                args = (rule, dry_run, par_updated, i in main_ids, callback)

                if sendable[i]:
                    res = process_rule(*args, pool)
                else:
                    res = process_rule(*args, None)
            except (Exception, KeyboardInterrupt) as e:
                traceback.print_exc()
                callback(events.FatalError(rule, e))
            finally:
                set_result(i, res)

                if res is None:
                    return


def _test_interproc_portabability(
    objs: Sequence[object], ctx: SpawnContext
) -> List[bool]:
    n = len(objs)
    picklable = [True] * n

    sys.stderr.write("Checking inter-process transferability of rules\n")

    with ctx.Pool(1) as pool:
        for i, obj in enumerate(objs):
            try:
                pool.apply(_dummy_func, (obj,))
            except Exception:
                picklable[i] = False

    return picklable


def _dummy_func(_: object) -> bool:
    ...


def _log_sendable_stats(sendables: Sequence[bool]):
    n = len(sendables)
    ok = sum(1 for x in sendables if x)
    ng = n - ok

    if ng > 0:
        sys.stderr.write(
            f"{ng} of {n} Rules will be executed using threads in the main "
            "process instad of using multi-processing. This is because "
            "their method/args/kwargs contain some objects that cannot be "
            "transfered to child processes.\n"
        )



import os
import sys
import shutil
import subprocess
from html import escape
from pathlib import Path
from typing import Dict, List, Literal, Optional, Union

from typing_extensions import TypeAlias

from .core import IGroup, IRule
from ..logwriter import term_is_jupyter

StrOrPath: TypeAlias = "Union[str, os.PathLike[str]]"

RankDir = Literal["TD", "LR"]


def print_graphviz(
    group: IGroup,
    output_file: Optional[StrOrPath] = None,
    *,
    rankdir: RankDir = "LR",
):
    """Visualize the dependency graph using Graphviz
    Args:
        group (Group): Group node whose Rules will be visualized
        output_file (str|os.PathLike|None):
            If specified, graph will be written into the file.
            Otherwise, graph will be printed to the terminal (available on
            Jupyter only). Graph format depends on the file extension:

            - .svg: SVG
            - .htm or .html: HTML (SVG image embedded)
            - .dot: Graphviz's DOT code (text)
    """
    if output_file is None:
        if term_is_jupyter():
            from IPython.display import display, SVG  # type: ignore

            dot_code = gen_dot_code(group, rankdir=rankdir)
            svg = convert(dot_code, "svg").decode()
            display(SVG(svg))
            return
        else:
            raise Exception("Printing to console is available on Jupyter only")
    else:
        assert isinstance(output_file, (str, os.PathLike))
        output_file = str(output_file)

        dot_code = gen_dot_code(
            group, Path(output_file).parent, rankdir=rankdir
        )

        if output_file[-4:] == ".svg":
            data = convert(dot_code, "svg")
        elif output_file[-4:] == ".dot":
            data = dot_code.encode()
        elif output_file[-4:] == ".htm" or output_file[-5:] == ".html":
            data = convert(dot_code, "svg").decode()
            data = (
                '<!DOCTYPE html><html><head><meta charset="utf-8">'
                f"<title>graph</title></head><body>{data}</body></html>"
            )
            data = data.encode("utf8")
        else:
            raise ValueError(
                "Output file's extension must be .svg, .dot, .htm, or .html"
            )

        with open(output_file, "wb") as f:
            f.write(data)


def gen_dot_code(
    group: IGroup, basedir: Optional[StrOrPath] = None, rankdir: RankDir = "LR"
):
    if not isinstance(
        group, IGroup
    ):  # pyright: ignore [reportUnnecessaryIsInstance]
        raise TypeError("argument group must be Group")

    if rankdir not in {"TB", "LR"}:
        raise ValueError(f"rankdir must be TB or LR. Given: {rankdir}")

    gid: Dict[IGroup, int] = {}
    rid: Dict[IRule, int] = {}
    fid: Dict[str, int] = {}

    res: List[str] = []
    res.append("digraph {")
    res.append("  compound=true;")
    res.append(f"  rankdir={rankdir};")

    def rec_group(g: IGroup, idt: str, par_prefix: str):
        gid[g] = len(gid)

        name = "<ROOT>" if len(g.name_tuple) == 0 else g.name_tuple[-1]

        if par_prefix == "":
            prefix = g.prefix
        elif g.prefix[: len(par_prefix)] == par_prefix:
            prefix = "... " + g.prefix[len(par_prefix) :]
        else:
            prefix = g.prefix

        res.append(idt + f"subgraph cluster{gid[g]} {{")
        res.append(
            idt + f"  label = <<B>{escape(name)}</B> "
            f'(<FONT FACE="monospace">{escape(prefix)}</FONT>)>;'
        )
        res.append(idt + '  style = "rounded";')

        for child_group in g.groups.values():
            rec_group(child_group, idt + "  ", g.prefix)

        for name, child_rule in g.rules.items():
            proc_rulew(child_rule, name, idt + "  ", g.prefix)

        res.append(idt + "};")

    def proc_rulew(r: IRule, name: str, idt: str, par_prefix: str):
        rid[r] = len(rid)

        res.append(idt + f"subgraph cluster_r_{rid[r]} {{")
        res.append(idt + f"  label=<<B>{escape(name)}</B>>;")
        res.append(idt + '  bgcolor = "#E0FFFF";')

        par_prefix = os.path.abspath(par_prefix + "_")[:-1]

        for yf in r.files.values():
            fid[os.path.abspath(yf)] = len(fid)

            p = os.path.abspath(yf)
            if par_prefix != "" and p[: len(par_prefix)] == par_prefix:
                p = "... " + p[len(par_prefix) :]
            else:
                p = str(yf)

            res.append(
                idt + f"  f{fid[os.path.abspath(yf)]} ["
                f'label=<<FONT FACE="monospace">{escape(p)}</FONT>>; '
                f"style=filled; "
                f"color=white; "
                f"shape=plain; "
                f'margin="0.1,0.1"; '
                f'URL="{_mk_link(yf, basedir)}"; '
                f"];"
            )
        res.append(idt + "}")

    rec_group(group, "  ", "")

    for r in rid.keys():
        f0 = os.path.abspath(next(iter(r.files.values())))
        for xf in r.xfiles:
            if xf not in fid:
                fid[xf] = len(fid)
                res.append(
                    f"  f{fid[xf]} ["
                    f'label=<<FONT FACE="monospace">'
                    f"{escape(str(xf))}</FONT>>; "
                    f"shape=plain; "
                    f'URL="{_mk_link(xf, basedir)}"; '
                    f"];"
                )

            res.append(
                f"  f{fid[xf]} -> f{fid[f0]} " f"[lhead=cluster_r_{rid[r]}];"
            )

    res.append("}")

    return "\n".join(res) + "\n"


def convert(dot_code: str, t: str = "svg"):
    if shutil.which("dot") is None:
        raise Exception(
            "Graphviz is required. dot executable was not found in PATH."
        )

    p = subprocess.run(
        ["dot", f"-T{t}"],
        input=dot_code.encode(),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    if p.returncode != 0:
        sys.stderr.write(p.stderr.decode())
        raise Exception(
            f"Failed to create graph. dot exit with code {p.returncode}"
        )

    return p.stdout


def save_to_file(dot_code: str, fname: StrOrPath, t: str = "svg"):
    with open(fname, "wb") as f:
        f.write(convert(dot_code, t))


def _mk_link(p: StrOrPath, basedir: Optional[StrOrPath]) -> str:
    basedir = basedir or os.getcwd()

    try:
        return os.path.relpath(p, basedir)
    except Exception:
        pass

    return str(p)



from __future__ import annotations
import re
from abc import ABCMeta
from typing import List, Optional, Sequence, Literal, TypeVar, Union, Tuple, Any

from ..core import IGroup, IRule, IFile

SelectorKind = Literal["group", "rule", "file"]
SELECTOR_KINDS: Tuple[Literal["group"], Literal["rule"], Literal["file"]] = (
    "group",
    "rule",
    "file",
)

SEP = ";"


def get_offspring_groups(
    root: IGroup, dst: Optional[List[IGroup]] = None
) -> List[IGroup]:
    if dst is None:
        dst = []

    dst.append(root)

    for c in root.groups.values():
        get_offspring_groups(c, dst)

    return dst


T = TypeVar("T", IGroup, IRule, IFile)


class SelectorMixin(IGroup, metaclass=ABCMeta):
    def _select(self, pattern: Sequence[str], kind: SelectorKind) -> List[Any]:
        rxs: List[str] = []

        for p in pattern:
            assert len(p) > 0

            if p.find("**") != -1 and p != "**":
                raise ValueError(
                    'Invalid pattern: "**" can only be an entire component'
                )
            if p == "**":
                rxs.append(f"({SEP}[^{SEP}]+)*")
            else:
                if p == "*":
                    # single * does not match an empty str
                    rxs.append(f"{SEP}[^{SEP}]+")
                else:

                    def _repl(m: re.Match[str]) -> str:
                        x = m.group()
                        return f"[^{SEP}]*" if x == "*" else re.escape(x)

                    p = re.sub(r"\*|[^*]+", _repl, p)
                    rxs.append(f"{SEP}{p}")

        regex = re.compile("^" + "".join(rxs) + "$")

        offspring_groups = get_offspring_groups(self)

        def _search(
            targets: List[T], target_names: List[Tuple[str, ...]]
        ) -> List[T]:
            res: List[T] = []

            for target, target_name in zip(targets, target_names):
                target_name = target_name[len(self.name_tuple) :]
                if regex.match("".join(SEP + n for n in target_name)):
                    res.append(target)

            return res

        if kind == "group":
            target_names = [n.name_tuple for n in offspring_groups]
            return _search(offspring_groups, target_names)
        elif kind == "rule":
            target_rules: List[IRule] = []
            target_names: List[Tuple[str, ...]] = []
            for g in offspring_groups:
                for r in g.rules.values():
                    target_rules.append(r)
                    target_names.append(r.name_tuple)
            return _search(target_rules, target_names)
        elif kind == "file":
            target_files: List[IFile] = []
            target_names: List[Tuple[str, ...]] = []
            for g in offspring_groups:
                for r in g.rules.values():
                    for k, f in r.files.items():
                        target_files.append(f)
                        target_names.append((*r.name_tuple, k))  # optimize?
            return _search(target_files, target_names)
        else:
            raise Exception("unreachable")

    def select_rules(self, pattern: Union[str, Sequence[str]]) -> List[IRule]:
        return self._select(_parse_args_pattern(pattern), "rule")

    def select_files(self, pattern: Union[str, Sequence[str]]) -> List[IFile]:
        return self._select(_parse_args_pattern(pattern), "file")

    def select_groups(self, pattern: Union[str, Sequence[str]]) -> List[IGroup]:
        """Obtain child groups or rules of this group.

        Signatures:

            1. `select(group_tree_pattern: str)`
            2. `select(group_tree_pattern: Sequence[str], group:bool=False)`

        Args for Signature 1:
            group_tree_pattern (str):
                Pattern of the relative name of child nodes of this group.
                Pattern consists of names concatenated with the delimiter '/'.
                Double star '**' can appear as a name indicating zero or
                more repetition of arbitrary names.

                Single star can appear as a part of a name indicating zero
                or more repetition of arbitrary character.

                If `group_tree_pattern[-1] == '/'`, it matches groups only.
                Otherwise, it matches rules only.

                For example, calling g.select(pattern) with a pattern

                * `"a/b"  matches a rule `g.a.b`
                * "a/b/" matches a group `g.a.b`
                * "a*"   matches rules `g.a`, `g.a1`, `g.a2`, etc
                * "a*/"  matches groups `g.a`, `g.a1`, `g.a2`, etc
                * `"**"`   matches all the offspring rules of `g`
                * `"**/"`  matches all the offspring groups of `g`
                * `"a/**"` matches all the offspring rules of the group `g.a`
                * `"**/b"` matches all the offspring rules of `g` of name "b"

            group: ignored

        Args for Signature-2:
            group_tree_pattern (list[str] | tuple[str]):
                Pattern representation using a sequence of names.

                Following two are equivalent:

                * `g.select(["a", "*", "c", "**"])`
                * `g.select("a/*/c/**")`

                Following two are equivalent:

                * `g.select(["a", "*", "c", "**"], True)`
                * `g.select("a/*/c/**/")`

            group (bool):
                if False (default), select rules only.
                if True, select groups only.

        Returns:
            list[RuleNodeLike]|list[Group]: rule nodes or group nodes.

            * called with Signature-1 and pattern[-1] != '/' or
            * called with Signature-2 and group is False

        Note:
            Cases where Signature-2 is absolutely necessary is when you need
            to select a node whose name contains "/".
            For example, ::

                g = create_group('group')

                # this rule's name is "dir/a.txt"
                rule = g.add('dir/a.txt', func)

                g.select(['dir/a.txt']) == [rule]  # OK
                g.select('dir/a.txt') != []  # trying to match g['dir']['a.txt']
        """
        return self._select(_parse_args_pattern(pattern), "group")


def _parse_args_pattern(pattern: object) -> List[str]:
    if isinstance(pattern, str):
        if len(pattern) == 0:
            raise ValueError("pattern must not be an empty str")

        pattern = pattern.strip("/")
        return re.split("/+", pattern)
    elif isinstance(pattern, Sequence):
        if not all(
            isinstance(v, str)
            for v in pattern  # pyright: ignore [reportUnknownVariableType]
        ):
            raise TypeError("Pattern sequence items must be str")

        return list(pattern)  # pyright: ignore [reportUnknownVariableType]
    else:
        raise TypeError("Pattern must be str or sequence of str")



import os
import sys
import time
from abc import ABCMeta, abstractmethod
from pathlib import Path
from logging import Logger
from typing import (
    Callable,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    Sequence,
    List,
)

from ..atom import unwrap_memo_values

from ...memo import ILazyMemo, create_lazy_memo_type, StrHashMemo, IMemo

from ...core.make import MakeSummary

from .selector import get_offspring_groups

from ...logwriter import (
    Loglevel,
    WritableProtocol,
    HTMLJupyterWriter,
    IWriter,
    WritersWrapper,
    HTMLFileWriterOpenOnDemand,
    TextFileWriterOpenOnDemand,
    LoggerWriter,
    ColorTextWriter,
    TextWriter,
    typeguard_loglevel,
    term_is_jupyter,
)
from ..core import IGroup, GroupTreeInfo, make, parse_args_prefix
from ...utils.strpath import StrOrPath


T = TypeVar("T")


class BasicInitMixin(IGroup, metaclass=ABCMeta):
    def __init__(
        self,
        dirname: Optional[StrOrPath] = None,
        prefix: Optional[StrOrPath] = None,
        *,
        loglevel: Optional[Loglevel] = None,
        use_default_logger: bool = True,
        logfile: Union[
            None,
            StrOrPath,
            Logger,
            WritableProtocol,
            Sequence[Union[StrOrPath, Logger, WritableProtocol]],
        ] = None,
    ):
        writer = basic_init_create_logwriter(
            loglevel, use_default_logger, logfile
        )

        lazy_memo_type = create_lazy_memo_type(StrHashMemo.create)
        memo_factory = create_lazy_memo_factory(lazy_memo_type)

        info = GroupTreeInfo(writer, memo_factory)

        self.__init_as_child__(info, self, ())

        self.set_prefix(prefix=parse_args_prefix(dirname, prefix))

    @abstractmethod
    def __init_as_child__(
        self,
        info: GroupTreeInfo,
        parent: IGroup,
        name: Tuple[str, ...],
    ):
        ...


class BasicMixin(IGroup, metaclass=ABCMeta):
    def clean(self) -> None:
        for g in get_offspring_groups(self):
            for r in g.rules.values():
                r.clean()

    def touch(
        self,
        file: bool = True,
        memo: bool = True,
        create: bool = True,
        t: Optional[float] = None,
    ) -> None:
        if t is None:
            t = time.time()

        for g in get_offspring_groups(self):
            for r in g.rules.values():
                r.touch(file, memo, create, t)

    def make(
        self,
        dry_run: bool = False,
        keep_going: bool = False,
        *,
        njobs: Optional[int] = None,
    ) -> MakeSummary:
        """Make rules in this group and their dependencies

        Args:
            dry_run (bool):
                instead of actually excuting the methods,
                print expected execution logs.
            keep_going (bool):
                If False (default), stop everything when a rule fails.
                If True, when a rule fails, keep executing other rules
                except the ones depend on the failed rule.
            njobs (int):
                Maximum number of rules that can be made concurrently.
                Defaults to 1 (single process, single thread).

        See also:
            See the description of jtcmake.make for more detail of njobs
        """
        return make(
            self,
            dry_run=dry_run,
            keep_going=keep_going,
            njobs=njobs,
        )


def basic_init_create_logwriter(
    loglevel: object, use_default_logger: object, logfile: object
) -> IWriter:
    loglevel = loglevel or "info"

    if not typeguard_loglevel(loglevel):
        raise TypeError(f"loglevel must be {Loglevel}. Given {loglevel}")

    if not (isinstance(use_default_logger, bool) or use_default_logger is None):
        raise TypeError(
            f"use_default_logger must be bool or None. "
            f"Given {use_default_logger}"
        )

    writers = create_logwriter_list(loglevel, logfile)

    if use_default_logger:
        writers.append(create_default_logwriter(loglevel))

    return WritersWrapper(writers)


def create_logwriter_list(loglevel: Loglevel, logfile: object) -> List[IWriter]:
    if logfile is None:
        return []
    elif isinstance(logfile, (list, tuple)):
        logfile_: Sequence[object] = logfile
        return [create_logwriter(f, loglevel) for f in logfile_]
    else:
        return [create_logwriter(logfile, loglevel)]


def create_logwriter(f: object, loglevel: Loglevel) -> IWriter:
    """
    Args:
        f (str|PathLike|Logger|WritableProtocol): logging destination
    """
    if isinstance(f, (str, os.PathLike)):
        if Path(f).suffix == ".html":
            return HTMLFileWriterOpenOnDemand(loglevel, f)
        else:
            return TextFileWriterOpenOnDemand(loglevel, f)

    if isinstance(f, Logger):
        return LoggerWriter(f)

    if isinstance(f, WritableProtocol):
        _isatty = getattr(f, "isatty", None)
        if callable(_isatty) and _isatty():
            return ColorTextWriter(f, loglevel)

        return TextWriter(f, loglevel)

    raise TypeError(
        "Logging target must be either str (file name), os.PathLike, "
        "logging.Logger, or and object with `write` method. "
        f"Given {f}"
    )


def create_default_logwriter(loglevel: Loglevel) -> IWriter:
    if term_is_jupyter():
        return HTMLJupyterWriter(loglevel, os.getcwd())
    elif sys.stderr.isatty():
        return ColorTextWriter(sys.stderr, loglevel)
    else:
        return TextWriter(sys.stderr, loglevel)


def create_lazy_memo_factory(
    lazy_memo_type: Type[ILazyMemo],
) -> Callable[[object], IMemo]:
    def _res(args: object) -> IMemo:
        args, lazy_args = unwrap_memo_values(args)
        return lazy_memo_type.create(args, lazy_args)

    return _res



from __future__ import annotations
from os import PathLike, fspath
from typing import Union
from typing_extensions import TypeAlias

StrOrPath: TypeAlias = "Union[str, PathLike[str]]"


def fspath2str(p: object) -> str:
    if isinstance(p, str):
        return p
    elif isinstance(p, PathLike):
        s = fspath(p)  # pyright: ignore [reportUnknownVariableType]
        if isinstance(s, str):
            return s
        elif isinstance(s, bytes):
            raise TypeError(f"bytes path is not supported. (given {s})")

    raise TypeError(f"Invalid path {p}")



from multiprocessing.pool import Pool
import traceback
import enum
from typing import Callable, List, NamedTuple, Optional, Set, Sequence, TypeVar

from . import events
from .abc import IRule, IEvent, UpdateResults


class Result(enum.Enum):
    Update = 1
    Skip = 2
    Fail = 3
    Fatal = 4


class MakeSummary(NamedTuple):
    total: int  # planned to be update
    update: int  # actually updated (method called)
    skip: int  # "nothing to do with ..."
    fail: int  # failed ones
    discard: int  # not checked because of abort


def _toplogical_sort(
    id2rule: Sequence[IRule], seed_ids: Sequence[int]
) -> List[int]:
    added: Set[int] = set()
    res: List[int] = []

    def rec(i: int):
        if i in added:
            return

        added.add(i)

        for dep in id2rule[i].deps:
            rec(dep)

        res.append(i)

    for i in seed_ids:
        rec(i)

    return res


_T_Rule = TypeVar("_T_Rule", bound=IRule)


def make(
    id2rule: Sequence[_T_Rule],
    ids: Sequence[int],
    dry_run: bool,
    keep_going: bool,
    callback: Callable[[IEvent[_T_Rule]], None],
):
    if len(ids) == 0:
        return MakeSummary(total=0, update=0, skip=0, fail=0, discard=0)

    main_ids = set(ids)

    taskq = _toplogical_sort(id2rule, ids)

    failed_ids: Set[int] = set()  # includes discarded ones
    updated_ids: Set[int] = set()
    nskips: int = 0  # for stats report
    nfails: int = 0  # for stats report

    for i in taskq:
        r = id2rule[i]

        if any(dep in failed_ids for dep in r.deps):
            failed_ids.add(i)
            continue

        par_updated = any(dep in updated_ids for dep in r.deps)

        try:
            result = process_rule(
                r, dry_run, par_updated, i in main_ids, callback, None
            )
        except Exception as e:
            result = Result.Fatal
            try:
                callback(events.FatalError(r, e))
            except Exception:
                traceback.print_exc()
                pass

        if result == Result.Update:
            updated_ids.add(i)
        elif result == Result.Fail:
            nfails += 1
            failed_ids.add(i)
            if not keep_going:
                try:
                    callback(events.StopOnFail())
                except Exception:
                    traceback.print_exc()
                break
        elif result == Result.Fatal:
            nfails += 1
            failed_ids.add(i)
            break
        else:
            assert result == Result.Skip
            nskips += 1

    return MakeSummary(
        total=len(taskq),
        update=len(updated_ids),
        skip=nskips,
        fail=nfails,
        discard=len(taskq) - (len(updated_ids) + nskips + nfails),
    )


def process_rule(
    rule: _T_Rule,
    dry_run: bool,
    par_updated: bool,
    is_main: bool,
    callback: Callable[[IEvent[_T_Rule]], None],
    pool: Optional[Pool],
):
    if dry_run:
        res = rule.check_update(par_updated, True)

        if isinstance(res, UpdateResults.Infeasible):
            callback(events.UpdateInfeasible(rule, res.reason))
            return Result.Fail
        elif isinstance(
            res, (UpdateResults.Necessary, UpdateResults.PossiblyNecessary)
        ):
            callback(events.DryRun(rule))
            return Result.Update
        else:
            callback(events.Skip(rule, is_main))
            return Result.Skip

    res = rule.check_update(par_updated, False)

    if isinstance(res, UpdateResults.Infeasible):
        callback(events.UpdateInfeasible(rule, res.reason))
        return Result.Fail
    elif isinstance(res, UpdateResults.UpToDate):
        callback(events.Skip(rule, is_main))
        return Result.Skip
    elif not isinstance(res, UpdateResults.Necessary):
        raise Exception(f"Internal error: unexpected result {res}")

    callback(events.Start(rule))

    try:
        rule.preprocess()
    except Exception as e:
        callback(events.PreProcError(rule, e))
        return Result.Fail

    try:
        if pool is None:
            rule.method()
            succ = True
        else:
            pool.apply(rule.method)
            succ = True
    except Exception as e:
        callback(events.ExecError(rule, e))
        succ = False
    except KeyboardInterrupt:
        try:
            rule.postprocess(False)
        except Exception:
            pass

        raise KeyboardInterrupt()

    try:
        rule.postprocess(succ)
    except Exception as e:
        callback(events.PostProcError(rule, e))
        return Result.Fail

    if succ:
        callback(events.Done(rule))
        return Result.Update
    else:
        return Result.Fail



from __future__ import annotations
import os
import inspect
import time
import re
from os import PathLike
from typing import (
    Any,
    Callable,
    Mapping,
    Optional,
    Tuple,
    Dict,
    Generic,
    Union,
    Sequence,
    List,
    TypeVar,
    Collection,
    Container,
    overload,
)
from typing_extensions import TypeGuard, ParamSpec, Concatenate

from ..core.make import MakeSummary

from ..utils.dict_view import DictView
from ..utils.nest import map_structure
from ..utils.strpath import StrOrPath
from .atom import unwrap_real_values
from .file import File, VFile
from .core import (
    IRule,
    GroupTreeInfo,
    IGroup,
    concat_prefix,
    require_tree_init,
    make,
    IFile,
    IAtom,
)
from .event_logger import INoArgFunc

K = TypeVar("K", bound=str)
P = ParamSpec("P")
T = TypeVar("T")


class _NoArgFunc(INoArgFunc):
    __slots__ = ("_method", "_args", "_kwargs")

    def __init__(
        self,
        method: Callable[..., object],
        args: Tuple[object, ...],
        kwargs: Dict[str, object],
    ) -> None:
        self._method = method
        self._args = args
        self._kwargs = kwargs

    @property
    def method(self) -> Callable[..., object]:
        return self._method

    @property
    def args(self) -> Tuple[object, ...]:
        return self._args

    @property
    def kwargs(self) -> Dict[str, object]:
        return self._kwargs


class SelfRule:
    __slots__ = ["_key"]
    _key: Union[str, int, None]

    def __init__(self, key: Union[None, str, int] = None):
        self._key = key

    def __getitem__(self, key: Union[str, int]) -> SelfRule:
        return SelfRule(key)

    def __getattr__(self, key: str) -> SelfRule:
        return self[key]

    def __repr__(self):
        if self._key is None:
            return self.__class__.__name__
        else:
            return f"{self.__class__.__name__}({self._key})"

    @property
    def key(self) -> Union[None, str, int]:
        return self._key


SELF: Any = SelfRule()

_T_Rule = TypeVar("_T_Rule", bound=IRule)
_T_deco_f = TypeVar("_T_deco_f", bound=Callable[[], object])


def require_init(
    rule_method: Callable[Concatenate[_T_Rule, P], T]
) -> Callable[Concatenate[_T_Rule, P], T]:
    def _method(self: _T_Rule, *args: P.args, **kwargs: P.kwargs) -> T:
        info = self._get_info()  # pyright: ignore [reportPrivateUsage]

        if self.name_tuple in info.rules_to_be_init:
            try:
                method_name = rule_method.__name__
            except AttributeError:
                method_name = "<unknown method>"

            raise Exception(
                f"Rule {self.name} must be initialized "
                f"before calling the method ({method_name})"
            )

        return rule_method(self, *args, **kwargs)

    return _method


class Rule(IRule, Generic[K]):
    _raw_rule_id: int
    _info: GroupTreeInfo
    _name: Tuple[str]
    _files: DictView[K, IFile]
    _xfiles: Sequence[str]
    _file_keys_hint: Optional[List[K]]
    _file_keys: List[K]
    _parent: IGroup

    def __init_partial__(
        self,
        name: Tuple[str, ...],
        group_tree_info: GroupTreeInfo,
        file_keys_hint: Optional[List[K]],
        parent: IGroup,
    ):
        self._info = group_tree_info
        self._name = name
        self._file_keys_hint = file_keys_hint
        self._parent = parent
        self._file_keys = []  # do not read before the rule is fully initialized

        self._info.rules_to_be_init.add(name)

    def __init_full__(
        self,
        yfiles: Dict[K, IFile],
        method: object,
        args: Tuple[object, ...],
        kwargs: Dict[str, object],
    ):
        if self.initialized:
            raise RuntimeError(f"Rule {self.name} is already initialized")

        self._init_main(yfiles, method, args, kwargs)
        self._info.rules_to_be_init.remove(self._name)

    def __init_at_once__(
        self,
        name: Tuple[str, ...],
        group_tree_info: GroupTreeInfo,
        parent: IGroup,
        yfiles: Mapping[K, IFile],
        method: object,
        args: Tuple[object, ...],
        kwargs: Dict[str, object],
    ) -> Rule[K]:
        self._info = group_tree_info
        self._name = name
        self._file_keys_hint = None
        self._parent = parent
        self._file_keys = []  # do not read before the rule is fully initialized
        self._init_main(yfiles, method, args, kwargs)

        return self

    @property
    def initialized(self) -> bool:
        return self._name not in self._info.rules_to_be_init

    def __getattr__(self, key: K) -> IFile:
        """Just for type checkers"""
        raise AttributeError(key)

    def __getitem__(self, key: Union[int, K]) -> IFile:
        if isinstance(key, int):
            return self._files[self._file_keys[key]]
        else:
            return self._files[key]

    @property
    def parent(self) -> IGroup:
        return self._parent

    def touch(
        self,
        file: bool = True,
        memo: bool = True,
        create: bool = True,
        t: Union[float, None] = None,
    ) -> None:
        logwriter = self._info.logwriter

        if t is None:
            t = time.time()

        if file:
            for f in self.files.values():
                if not f.exists():
                    if not create:
                        continue

                    f.touch()

                os.utime(f, (t, t))
                logwriter.info(f"touch {f}")

        if memo:
            self._info.rule_store.rules[self.raw_rule_id].update_memo()

    def clean(self) -> None:
        logwriter = self._info.logwriter

        for f in self.files.values():
            if not f.exists():
                continue

            try:
                f.unlink(missing_ok=False)
            except Exception as e:
                logwriter.warning(f"Failed to remove {f}. {e}")
            else:
                logwriter.info(f"Delete {f}")

    def _init_main(
        self,
        yfiles: Mapping[K, IFile],
        method: object,
        args: object,
        kwargs: object,
    ):
        args_ = (args, kwargs)

        # Add path prefix
        yfiles = {
            k: type(f)(_normalize_path(str(f), self._parent.prefix))
            for k, f in yfiles.items()
        }

        # (Abspath of output) => (IFile of output)
        yp2f = {os.path.abspath(f): f for f in yfiles.values()}

        # Replace reserved objects by Atoms
        args_ = _replace_obj_by_atom_in_structure(self._info.memo_store, args_)

        # Replace SELFs
        args_ = _replace_self({k: v for k, v in yfiles.items()}, args_)

        # Assert that all outputs are included in the arguments
        _assert_all_yfiles_used_in_args(yp2f, args_)

        # (Abspath of input) => (IFile of input)
        xp2f = _find_xfiles_in_args(yp2f, args_)

        # Create final method arguments
        method_args, method_kwargs = unwrap_real_values(args_)  # type: ignore

        # Validate method signature
        if not callable(method):
            raise TypeError(f"method must be callable. Given {method}")

        _assert_signature_match(method, method_args, method_kwargs)
        method_ = _NoArgFunc(method, method_args, method_kwargs)

        # Create memo
        memo = self._info.memo_factory(args_)

        # Update the RuleStore (create and add a new raw Rule)
        raw_rule = self._info.rule_store.add(
            yp2f, xp2f, method_, memo, self._name
        )

        # Update self
        self._raw_rule_id = raw_rule.id
        self._files = DictView(yfiles)
        self._xfiles = list(xp2f)
        self._file_keys = list(yfiles)

        for k, f in yfiles.items():
            if k.isidentifier() and not hasattr(self, k):
                setattr(self, k, f)

    @overload
    def init(self, method: Callable[P, None], /) -> Callable[P, Rule[K]]:
        ...

    @overload
    def init(
        self,
        output_files: Union[
            Mapping[K, StrOrPath],
            Sequence[Union[K, PathLike[K]]],
            K,
            PathLike[K],
        ],
        method: Callable[P, object],
        /,
    ) -> Callable[P, Rule[K]]:
        ...

    def init(
        self, output_files: object, method: object = None, /
    ) -> Callable[..., Rule[K]]:
        return self._init(output_files, method, IFile_fact=File)

    @overload
    def initvf(self, method: Callable[P, None], /) -> Callable[P, Rule[K]]:
        ...

    @overload
    def initvf(
        self,
        output_files: Union[
            Mapping[K, StrOrPath],
            Sequence[Union[K, PathLike[K]]],
            K,
            PathLike[K],
        ],
        method: Callable[P, object],
        /,
    ) -> Callable[P, Rule[K]]:
        ...

    def initvf(
        self, output_files: object, method: object = None, /
    ) -> Callable[..., Rule[K]]:
        return self._init(output_files, method, IFile_fact=VFile)

    def _init(
        self,
        output_files: object,
        method: object = None,
        /,
        *,
        IFile_fact: Callable[[StrOrPath], IFile],
    ) -> Callable[P, Rule[K]]:
        if method is None:
            output_files, method = self.name_tuple[-1], output_files

        yfiles = parse_args_output_files(
            self.name_tuple[-1], self._file_keys_hint, output_files, IFile_fact
        )

        def _init(*args: P.args, **kwargs: P.kwargs) -> Rule[K]:
            self.__init_full__(yfiles, method, args, kwargs)
            return self

        return _init

    def init_deco(
        self,
        output_files: Optional[
            Union[
                Mapping[K, StrOrPath],
                Sequence[Union[K, PathLike[K]]],
                K,
                PathLike[K],
            ]
        ] = None,
    ) -> Callable[[_T_deco_f], _T_deco_f]:
        return self._init_deco(output_files, IFile_fact=File)

    def initvf_deco(
        self,
        output_files: Optional[
            Union[
                Mapping[K, StrOrPath],
                Sequence[Union[K, PathLike[K]]],
                K,
                PathLike[K],
            ]
        ] = None,
    ) -> Callable[[_T_deco_f], _T_deco_f]:
        return self._init_deco(output_files, IFile_fact=VFile)

    def _init_deco(
        self,
        output_files: object = None,
        *,
        IFile_fact: Callable[[StrOrPath], IFile],
    ) -> Callable[[_T_deco_f], _T_deco_f]:
        if output_files is None:
            output_files = self.name_tuple[-1]

        yfiles = parse_args_output_files(
            self.name_tuple[-1], self._file_keys_hint, output_files, IFile_fact
        )

        def decorator(method: _T_deco_f):
            args, kwargs = Rule_init_parse_deco_func(method)
            self.__init_full__(yfiles, method, args, kwargs)
            return method

        return decorator

    def _get_info(self) -> GroupTreeInfo:
        return self._info

    @property
    @require_init
    def raw_rule_id(self) -> int:
        return self._raw_rule_id

    @require_tree_init
    def make(
        self,
        dry_run: bool = False,
        keep_going: bool = False,
        *,
        njobs: Optional[int] = None,
    ) -> MakeSummary:
        """Make this rule and its dependencies
        Args:
            dry_run:
                instead of actually excuting the methods,
                print expected execution logs.
            keep_going:
                If False (default), stop everything when a rule fails.
                If True, when a rule fails, keep executing other rules
                except the ones depend on the failed rule.
            njobs:
                Maximum number of rules that can be made concurrently.
                Defaults to 1 (single process, single thread).

        See also:
            See the description of jtcmake.make for more detail of njobs
        """
        return make(
            self,
            dry_run=dry_run,
            keep_going=keep_going,
            njobs=njobs,
        )

    @property
    def name_tuple(self) -> Tuple[str, ...]:
        return self._name

    @property
    @require_init
    def files(self) -> Mapping[K, IFile]:
        return self._files

    @property
    @require_init
    def xfiles(self) -> Collection[str]:
        return self._xfiles


def Rule_init_parse_deco_func(
    method: object,
) -> Tuple[Tuple[object, ...], Dict[str, object]]:
    if not callable(method):
        raise TypeError(f"method must be callable. Given {method}")

    sig = inspect.signature(method)
    params = sig.parameters

    nodefaults = [
        name
        for name, p in params.items()
        if p.default is inspect.Parameter.empty
    ]

    if len(nodefaults) != 0:
        raise TypeError(
            "All the arguments of the method must have a default value. "
            f"Missing ones are: {nodefaults}"
        )

    for name, p in params.items():
        if p.kind == inspect.Parameter.POSITIONAL_ONLY:
            raise TypeError(
                f"method must not have a positional only argument. "
                f"{name} is a positonal only argument."
            )

    kwargs = {k: v.default for k, v in params.items()}

    return (), kwargs


def _validate_str(s: object, err_msg: str) -> str:
    if not isinstance(s, str):
        raise TypeError(err_msg)

    return s


T_type = TypeVar("T_type")


def _validate_type(tp: Tuple[type[T_type], ...], o: object, msg: str) -> T_type:
    if isinstance(o, tp):
        return o  # pyright: ignore
    else:
        raise TypeError(msg)


def _pathlike_to_str(p: object) -> str:
    if not isinstance(p, (str, os.PathLike)):
        raise TypeError("Expected str or os.PathLike. Got {p}")

    return _validate_str(os.fspath(p), "bytes path is not allowed. ({p})")


def _to_IFile(f: object, IFile_factory: Callable[[StrOrPath], IFile]) -> IFile:
    if isinstance(f, IFile):
        return f

    if isinstance(f, (str, os.PathLike)):
        return IFile_factory(f)

    raise TypeError(f"Output file must be str or PathLike. Given {f}")


def parse_args_output_files(
    rule_name: str,
    key_hints: Optional[Sequence[K]],
    output_files: object,
    IFile_factory: Callable[[StrOrPath], IFile],
) -> Dict[K, IFile]:
    if isinstance(output_files, (tuple, list)):
        output_files_str = map(_pathlike_to_str, output_files)
        output_files_str = (
            _repl_name_ref(p, rule_name, None) for p in output_files_str
        )
        outs: Dict[str, IFile] = {
            v: type(_to_IFile(f, IFile_factory))(v)
            for v, f in zip(  # pyright: ignore [reportUnknownVariableType]
                output_files_str, output_files
            )
        }
    elif isinstance(output_files, (str, os.PathLike)):
        k = _pathlike_to_str(output_files)
        k = _repl_name_ref(k, rule_name, None)
        outs = {k: type(_to_IFile(output_files, IFile_factory))(k)}
    elif isinstance(output_files, Mapping):
        output_files_: Mapping[object, object] = output_files
        keys = [
            _validate_type((str,), k, "file key must be str")
            for k in output_files_
        ]
        files_str = map(_pathlike_to_str, output_files_.values())
        files_str = (
            _repl_name_ref(f, rule_name, k) for k, f in zip(keys, files_str)
        )

        files = (
            type(_to_IFile(f, IFile_factory))(sf)
            for f, sf in zip(output_files_.values(), files_str)
        )

        outs = {k: _to_IFile(f, IFile_factory) for k, f in zip(keys, files)}
    else:
        raise TypeError(
            "output_files must be str | PathLike | Sequence[str|PathLike] "
            "| Mapping[str|PathLike]."
        )

    if not _check_file_keys(outs, key_hints):
        raise TypeError(
            f"Runtime type check error: Expected file keys: {key_hints}, "
            f"Actual keys of given output files: {list(outs.keys())}"
        )

    return outs


def _check_file_keys(
    files: Dict[str, IFile], ref: Optional[Sequence[K]]
) -> TypeGuard[Dict[K, IFile]]:
    if ref is None:
        return True

    return len(files) == len(ref) and all(a == b for a, b in zip(files, ref))


def _normalize_path(p: str, pfx: str) -> str:
    p = concat_prefix(p, pfx)

    try:
        rel = os.path.relpath(p, os.getcwd())
        return rel if len(rel) < len(p) else p
    except Exception:
        return p


def _replace_obj_by_atom_in_structure(
    memo_store: Mapping[int, IAtom], args: object
) -> object:
    def _rec(o: object) -> object:
        if id(o) in memo_store:
            return memo_store[id(o)]
        elif isinstance(o, dict):
            _o: Dict[object, object] = o
            return {k: _rec(v) for k, v in _o.items()}
        elif isinstance(o, tuple):
            return tuple(map(_rec, o))
        elif isinstance(o, list):
            return list(map(_rec, o))
        elif isinstance(o, set):
            return set(map(_rec, o))
        else:
            return o

    return _rec(args)


def _replace_self(files: Mapping[str, IFile], args: object) -> object:
    def repl(o: object):
        if isinstance(o, SelfRule):
            if o.key is None:
                if len(files) >= 2:
                    raise TypeError(
                        "Self-without-key is not allowed when the "
                        "rule has multiple output files"
                    )

                return next(iter(files.values()))
            elif isinstance(o.key, int):
                if o.key >= len(files):
                    raise IndexError(
                        f"SELF index is {o.key} but the rule "
                        f"has only {len(files)} output files"
                    )
                return list(files.values())[o.key]
            else:
                if o.key not in files:
                    raise KeyError(f"Failed to resolve SELF: {o.key}")
                return files[o.key]
        else:
            return o

    return map_structure(repl, args)


def _assert_all_yfiles_used_in_args(ypaths: Collection[str], args: object):
    unused = set(ypaths)

    def check(v: object):
        if isinstance(v, IFile):
            absp = os.path.abspath((v))
            if absp in unused:
                unused.remove(absp)

    map_structure(check, args)

    if len(unused) > 0:
        raise ValueError(
            f"All the output files must appear in the method arguments. "
            f"Unused ones are: {unused}"
        )


def _find_xfiles_in_args(
    ypaths: Container[str], args: object
) -> Dict[str, IFile]:
    res: Dict[str, IFile] = {}

    def check(v: object):
        if isinstance(v, IFile):
            absp = os.path.abspath(v)
            if absp not in ypaths:
                res[absp] = v

    map_structure(check, args)

    return res


def _assert_signature_match(
    func: Callable[..., object],
    args: Sequence[object],
    kwargs: Dict[str, object],
):
    try:
        inspect.signature(func).bind(*args, **kwargs)
    except Exception as e:
        raise TypeError(
            "Signature of the method does not match the arguments"
        ) from e


NAME_REF_RULE = "<R>"
NAME_REF_FILE = "<F>"


def _repl_name_ref(src: str, rule_name: str, file_key: Optional[str]) -> str:
    def _repl(m: re.Match[str]) -> str:
        r = m.group(0)
        if r == NAME_REF_RULE:
            return rule_name
        else:
            if file_key is None:
                raise ValueError(
                    "File key may not referenced by the output file names "
                    "when the keys weren't explicitly given."
                )

            return file_key

    pattern = f"{re.escape(NAME_REF_RULE)}|{re.escape(NAME_REF_FILE)}"

    a = re.sub(pattern, _repl, src)
    return a



from __future__ import annotations
from typing import TypeVar
from traceback import TracebackException
from .abc import IEvent, IRule


_T_Rule = TypeVar("_T_Rule", bound=IRule)


class RuleEvent(IEvent[_T_Rule]):
    def __init__(self, rule: _T_Rule):
        self._rule = rule

    @property
    def rule(self) -> _T_Rule:
        return self._rule

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.rule})"


class ErrorRuleEvent(RuleEvent[_T_Rule]):
    def __init__(self, rule: _T_Rule, err: BaseException):
        """
        Args:
            err (BaseException):
                This will be immediately coverted to a lighter format which
                does not have references to the stack frames.
                The format is currently traceback.TracebackException,
                which the implementor expect has no reference to stack frames.
                Fix this class if the implementor's expectation is not true.
        """
        super().__init__(rule)
        self.trace_exc = TracebackException.from_exception(err)


class Skip(RuleEvent[_T_Rule]):
    def __init__(self, rule: _T_Rule, is_direct_target: bool):
        super().__init__(rule)
        self.is_direct_target = is_direct_target


class Start(RuleEvent[_T_Rule]):
    ...


class Done(RuleEvent[_T_Rule]):
    ...


class DryRun(RuleEvent[_T_Rule]):
    ...


class StopOnFail(IEvent[_T_Rule]):
    ...


class UpdateInfeasible(RuleEvent[_T_Rule]):
    def __init__(self, rule: _T_Rule, reason: str):
        super().__init__(rule)
        self.reason = reason


class PreProcError(ErrorRuleEvent[_T_Rule]):
    ...


class ExecError(ErrorRuleEvent[_T_Rule]):
    ...


class PostProcError(ErrorRuleEvent[_T_Rule]):
    ...


class FatalError(ErrorRuleEvent[_T_Rule]):
    ...



import abc
import os
import html
from logging import Logger
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
)

from typing_extensions import Literal, Protocol, runtime_checkable, TypeGuard

from .utils.strpath import StrOrPath

T = TypeVar("T")


def _first_non_none(*t: T) -> T:
    assert len(t) > 0
    for x in t:
        if x is not None:
            return x

    return t[-1]


class RichStrAttr(NamedTuple):
    c: Optional[Tuple[int, int, int]] = None
    bg: Optional[Tuple[int, int, int]] = None
    link: Optional[StrOrPath] = None


class RichStr(str):
    attr: RichStrAttr

    def __new__(cls, s: str, *_args: object, **_kwargs: object):
        return super().__new__(cls, s)

    def __init__(
        self,
        s: str,
        c: Optional[Tuple[int, int, int]] = None,
        bg: Optional[Tuple[int, int, int]] = None,
        link: Optional[StrOrPath] = None,
        default: Optional[RichStrAttr] = None,
    ):
        a = s.attr if isinstance(s, RichStr) else RichStrAttr()
        b = default or RichStrAttr()

        self.attr = RichStrAttr(
            _first_non_none(c, a.c, b.c),
            _first_non_none(bg, a.bg, b.bg),
            _first_non_none(link, a.link, b.link),
        )

    def __add__(self, rhs: object):
        if type(rhs) == str:
            return RichStr(str(self) + str(rhs), *self.attr)
        else:
            return NotImplemented

    def __radd__(self, lhs: object):
        if isinstance(lhs, str):
            return RichStr(str(lhs) + str(self), *self.attr)
        else:
            return NotImplemented


Loglevel = Literal["debug", "info", "warning", "error"]


def typeguard_loglevel(loglevel: object) -> TypeGuard[Loglevel]:
    return loglevel in {"debug", "info", "warning", "error"}


QUANT_LOG_LEVEL: Mapping[Loglevel, int] = {
    "debug": 10,
    "info": 20,
    "warning": 30,
    "error": 40,
}


class IWriter:
    loglevel: Loglevel

    def __init__(self, loglevel: Loglevel):
        assert loglevel in {"debug", "info", "warning", "error"}

        self.loglevel = loglevel

    @abc.abstractmethod
    def _write(self, *args: str, level: Loglevel):
        ...

    def write(self, *args: str, level: Loglevel):
        if QUANT_LOG_LEVEL[self.loglevel] <= QUANT_LOG_LEVEL[level]:
            self._write(*args, level=level)

    def debug(self, *args: str):
        self.write(*args, level="debug")

    def info(self, *args: str):
        self.write(*args, level="info")

    def warning(self, *args: str):
        self.write(*args, level="warning")

    def error(self, *args: str):
        self.write(*args, level="error")


class WritersWrapper(IWriter):
    def __init__(
        self, writers: Sequence[IWriter], loglevel: Optional[Loglevel] = None
    ):
        super().__init__(loglevel or "debug")
        assert all(isinstance(w, IWriter) for w in writers)
        self.writers = writers

    def _write(self, *args: str, level: Loglevel):
        for writer in self.writers:
            writer.write(*args, level=level)


@runtime_checkable
class WritableProtocol(Protocol):
    def write(self, __t: str) -> Any:
        ...


class TextWriter(IWriter):
    def __init__(self, writable: WritableProtocol, loglevel: Loglevel):
        super().__init__(loglevel)
        self.writable = writable

    def _write(self, *args: str, level: Loglevel):
        self.writable.write("".join(map(str, args)) + "\n")


class ColorTextWriter(IWriter):
    def __init__(self, writable: WritableProtocol, loglevel: Loglevel):
        super().__init__(loglevel)
        self.writable = writable

    def _write(self, *args: str, level: Loglevel):
        color = (
            {
                "debug": (0x4F, 0x8A, 0x10),
                "info": None,
                "warning": (0x9F, 0x60, 0x00),
                "error": (0xD8, 0x00, 0x0C),
            }
        ).get(level)

        bgcolor = (
            {
                "debug": None,
                "info": None,
                "warning": None,
                "error": None,
            }
        ).get(level)

        args_ = [RichStr(x, default=RichStrAttr(color, bgcolor)) for x in args]
        self.writable.write(create_color_str(args_) + "\n")


class LoggerWriter(IWriter):
    def __init__(self, logger: Logger):
        super().__init__("debug")
        self.logger = logger

    def _write(self, *args: str, level: Loglevel):
        msg = "".join(map(str, args))

        if level == "debug":
            self.logger.debug(msg)
        elif level == "info":
            self.logger.info(msg)
        elif level == "warning":
            self.logger.warning(msg)
        elif level == "error":
            self.logger.error(msg)


class TextFileWriterOpenOnDemand(IWriter):
    def __init__(self, loglevel: Loglevel, fname: StrOrPath):
        super().__init__(loglevel)

        os.makedirs(Path(fname).parent, exist_ok=True)

        self.fname = fname

    def _write(self, *args: str, level: Loglevel):
        with open(self.fname, "a") as f:
            f.write("".join(map(str, args)) + "\n")


HTML_BG_COLOR_MAP: Dict[Loglevel, Tuple[int, int, int]] = {
    "debug": (0xDF, 0xF2, 0xBF),
    "info": (0xFF, 0xFF, 0xFF),
    "warning": (0xFE, 0xEF, 0xB3),
    "error": (0xFF, 0xD2, 0xD2),
}
HTML_COLOR_MAP: Dict[Loglevel, Tuple[int, int, int]] = {
    "debug": (0x4F, 0x8A, 0x10),
    "info": (0, 0, 0),
    "warning": (0x9F, 0x60, 0x00),
    "error": (0xD8, 0x00, 0x0C),
}


class HTMLFileWriterOpenOnDemand(IWriter):
    basedir: Optional[Path]
    fname: Path

    def __init__(
        self,
        loglevel: Loglevel,
        fname: StrOrPath,
        basedir: Optional[Path] = None,
    ):
        super().__init__(loglevel)

        os.makedirs(Path(fname).parent, exist_ok=True)

        self.basedir = basedir
        self.fname = Path(fname)

    def _write(self, *args: str, level: Loglevel):
        color = HTML_COLOR_MAP[level]
        bgcolor = HTML_BG_COLOR_MAP[level]

        args_ = [RichStr(x, default=RichStrAttr(color, bgcolor)) for x in args]

        with open(self.fname, "a") as f:
            f.write(
                '<html><head><meta charset="utf-8"><title>log</title></head>'
                '<body><pre style="background-color: '
                f'rgb({bgcolor[0]}, {bgcolor[1]}, {bgcolor[2]})">'
            )
            f.write(create_html(args_, self.basedir))
            f.write("</pre></body></html>")


class HTMLJupyterWriter(IWriter):
    basedir: Optional[Path]

    def __init__(self, loglevel: Loglevel, basedir: Optional[StrOrPath] = None):
        super().__init__(loglevel)
        from IPython.display import display, HTML  # pyright: ignore

        del display, HTML

        self.basedir = Path(basedir) if basedir else None

    def _write(self, *args: str, level: Loglevel):
        from IPython.display import display, HTML  # pyright: ignore

        color = HTML_COLOR_MAP[level]
        bgcolor = HTML_BG_COLOR_MAP[level]

        args_ = [RichStr(x, default=RichStrAttr(color, bgcolor)) for x in args]

        display(
            HTML(
                '<pre style="background-color: '
                f'rgb({bgcolor[0]}, {bgcolor[1]}, {bgcolor[2]})">'
                f"{create_html(args_, self.basedir)}</pre>"
            )
        )


def create_html(sl: Sequence[str], basedir: Optional[StrOrPath] = None) -> str:
    sl = [x if isinstance(x, RichStr) else RichStr(x) for x in sl]
    groups: List[List[RichStr]] = []
    for s in sl:
        if len(groups) >= 1 and s.attr == groups[-1][0].attr:
            groups[-1].append(s)
        else:
            groups.append([s])

    outs: List[str] = []
    for group in groups:
        s = RichStr("".join(group), *group[0].attr)
        outs.append(_richstr_to_html(s, str(basedir)))

    return "".join(outs)


def _richstr_to_html(s: RichStr, basedir: str) -> str:
    starts: List[str] = []
    ends: List[str] = []

    attr = s.attr

    if attr.link is not None:
        if basedir is not None:
            try:
                link = os.path.relpath(attr.link, basedir)
            except Exception:
                link = attr.link
        else:
            link = attr.link
        starts.append(f'<a href="{Path(link).as_posix()}">')
        ends.append("</a>")

    styles: List[str] = []
    if attr.c is not None:
        styles.append(f"color: rgb({attr.c[0]}, {attr.c[1]}, {attr.c[2]});")

    if attr.bg is not None:
        styles.append(
            f"background-color: rgb({attr.bg[0]}, {attr.bg[1]}, {attr.bg[2]});"
        )

    if len(styles) != 0:
        style = "".join(styles)
        starts.append(f'<span style="{style}">')
        ends.append("</span>")

    starts.append(html.escape(s))
    starts.extend(ends[::-1])
    return "".join(starts)


def create_color_str(sl: Sequence[str]) -> str:
    res: List[str] = []
    last_c = None
    last_bg = None
    for s in sl:
        if not isinstance(s, RichStr):
            last_bg = None
            last_c = None
            res.append("\x1b[49m\x1b[39m" + s)
            continue

        attr = s.attr

        if attr.bg != last_bg:
            last_bg = attr.bg
            if attr.bg is None:
                res.append("\x1b[49m")
            else:
                res.append(f"\x1b[48;5;{_comp_8bit_term_color(*attr.bg)}m")

        if attr.c != last_c:
            last_c = attr.c
            if attr.c is None:
                res.append("\x1b[39m")
            else:
                res.append(f"\x1b[38;5;{_comp_8bit_term_color(*attr.c)}m")

        res.append(str(s))

    res.append("\x1b[49m\x1b[39m")

    return "".join(res)


def _comp_8bit_term_color(r: int, g: int, b: int) -> int:
    r, g, b = (x * 6 // 256 for x in (r, g, b))
    return 16 + r * 36 + g * 6 + b


def term_is_jupyter() -> bool:
    try:
        from IPython.core.getipython import get_ipython  # pyright: ignore

        name = get_ipython().__class__.__name__  # pyright: ignore
        return name == "ZMQInteractiveShell"  # pyright: ignore
    except Exception:
        return False



from __future__ import annotations
from abc import ABCMeta, abstractmethod
from typing import Callable, Union, Set, TypeVar, Generic
from typing_extensions import TypeAlias


class UpToDate:
    ...


class Necessary:
    ...


class PossiblyNecessary:
    # dry_run only
    ...


class Infeasible:
    def __init__(self, reason: str):
        """
        Args:
            reason (str): reason
        """
        self.reason = reason


class UpdateResults:
    UpToDate = UpToDate
    Necessary = Necessary
    PossiblyNecessary = PossiblyNecessary
    Infeasible = Infeasible


UpdateResult: TypeAlias = Union[
    UpToDate, Necessary, PossiblyNecessary, Infeasible
]


_T_Rule = TypeVar("_T_Rule", bound="IRule", covariant=True)


class IEvent(Generic[_T_Rule]):
    ...


class IRule(metaclass=ABCMeta):
    @abstractmethod
    def check_update(self, par_updated: bool, dry_run: bool) -> UpdateResult:
        ...

    @abstractmethod
    def preprocess(self) -> None:
        ...

    @abstractmethod
    def postprocess(self, succ: bool) -> None:
        ...

    @property
    @abstractmethod
    def method(self) -> Callable[[], object]:
        ...

    @property
    @abstractmethod
    def deps(self) -> Set[int]:
        ...



from __future__ import annotations
import os
import json
from hashlib import sha256
from abc import ABCMeta, abstractmethod
from numbers import Complex
from typing import (
    Callable,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    List,
    Union,
)

from .raw_rule import IMemo


ENCODING = "utf8"


class ILazyMemo(IMemo, metaclass=ABCMeta):
    @property
    def memo(self) -> IMemo:
        ...

    @property
    def lazy_memo(self) -> IMemo:
        ...

    @classmethod
    @abstractmethod
    def create(cls, args: object, lazy_args: Callable[[], object]) -> ILazyMemo:
        ...


def create_lazy_memo_type(
    memo_factory: Callable[[object], IMemo]
) -> Type[ILazyMemo]:
    from_bytes = memo_factory(None).loads

    class LazyMemo(ILazyMemo):
        __slots__ = ("_memo", "lazy_args")
        _memo: IMemo
        lazy_args: Union[Callable[[], object], IMemo]

        def __init__(
            self,
            memo: IMemo,
            lazy_args: Union[Callable[[], object], IMemo],
        ):
            self._memo = memo
            self.lazy_args = lazy_args

        @property
        def memo(self) -> IMemo:
            return self._memo

        @property
        def lazy_memo(self) -> IMemo:
            if isinstance(self.lazy_args, IMemo):
                return self.lazy_args
            else:
                return memo_factory(self.lazy_args())

        def compare(self, other: IMemo) -> bool:
            if isinstance(other, ILazyMemo):
                return self.memo.compare(other.memo) and self.lazy_memo.compare(
                    other.lazy_memo
                )
            else:
                return False

        @classmethod
        def create(
            cls, args: object, lazy_args: Callable[[], object]
        ) -> ILazyMemo:
            return cls(memo_factory(args), lazy_args)

        def dumps(self) -> Iterable[bytes]:
            memo_ = b"".join(self.memo.dumps()).hex()
            lazy_memo_ = b"".join(self.lazy_memo.dumps()).hex()
            return [json.dumps([memo_, lazy_memo_]).encode(ENCODING)]

        @classmethod
        def loads(cls, data: bytes) -> ILazyMemo:
            o = json.loads(data.decode(ENCODING))
            assert isinstance(o, list)

            memo = from_bytes(bytes.fromhex(o[0]))
            lazy_args = from_bytes(bytes.fromhex(o[1]))

            return cls(memo, lazy_args)

    return LazyMemo


MAX_RAW_REPRESENTATION_LEN = 1000


class StrHashMemo(IMemo):
    __slots__ = ("text",)
    text: str

    def __init__(self, text: str):
        if len(text) > MAX_RAW_REPRESENTATION_LEN:
            self.text = _hash_fn(text)
        else:
            self.text = text

    def compare(self, other: IMemo) -> bool:
        if isinstance(other, StrHashMemo):
            return self.text == other.text
        else:
            return False

    def dumps(self) -> Iterable[bytes]:
        return [self.text.encode(ENCODING)]

    @classmethod
    def loads(cls, data: bytes) -> StrHashMemo:
        return cls(data.decode(ENCODING))

    @classmethod
    def create(cls, args: object) -> StrHashMemo:
        return cls(stringify(args, None))


def _hash_fn(s: str):
    return sha256(s.encode("utf8")).digest().hex()


SUPPORTED_ATOM_TYPES = (  # pyright: ignore [reportUnknownVariableType]
    str,
    type(None),
    Complex,
    bool,
    bytes,
    bytearray,
    os.PathLike,
)


def stringify(
    nest: object, default: Optional[Callable[[object], object]]
) -> str:
    def _default(_: object):
        raise TypeError()

    dst: List[str] = []
    _stringify(nest, dst, set(), default or _default)

    return "".join(dst)


def _stringify(
    nest: object,
    dst: List[str],
    visited_container: Set[int],
    default: Callable[[object], object],
) -> None:
    if isinstance(nest, (tuple, list)):
        stringify_sequence(nest, dst, visited_container, default)
    elif isinstance(nest, set):
        stringify_set(nest, dst, visited_container, default)
    elif isinstance(nest, dict):
        stringify_mapping(nest, dst, visited_container, default)
    else:
        stringify_atom(nest, dst, visited_container, default)


def _check_and_update_visited(nest: object, visited_container: Set[int]):
    i = id(nest)
    if i in visited_container:
        raise Exception("Cannot serialize cyclic structure")
    visited_container.add(i)


def stringify_sequence(
    nest: Sequence[object],
    dst: List[str],
    visited_container: Set[int],
    default: Callable[[object], object],
):
    _check_and_update_visited(nest, visited_container)

    dst.append("(" if isinstance(nest, tuple) else "[")
    for v in nest:
        _stringify(v, dst, visited_container, default)
        dst.append(",")
    dst.append(")" if isinstance(nest, tuple) else "]")


def _stringify_keys(
    keys: Iterable[object],
    visited_container: Set[int],
    default: Callable[[object], object],
) -> List[Tuple[str, object]]:
    res: List[Tuple[str, object]] = []

    for k in keys:
        _s: List[str] = []
        _stringify(k, _s, visited_container, default)
        res.append(("".join(_s), k))

    return res


def stringify_mapping(
    nest: Mapping[object, object],
    dst: List[str],
    visited_container: Set[int],
    default: Callable[[object], object],
):
    _check_and_update_visited(nest, visited_container)

    skey_keys = _stringify_keys(nest, visited_container, default)

    try:
        skey_keys.sort()
    except TypeError as e:
        e_ = TypeError("dict in memoization values must have sortable keys")
        raise e_ from e

    dst.append("{")

    for sk, k in skey_keys:
        dst.append(sk)
        dst.append(":")
        _stringify(nest[k], dst, visited_container, default)
        dst.append(",")
    dst.append("}")


def stringify_set(
    nest: Set[object],
    dst: List[str],
    visited_container: Set[int],
    default: Callable[[object], object],
):
    _check_and_update_visited(nest, visited_container)

    skey_keys = _stringify_keys(nest, visited_container, default)

    try:
        skey_keys.sort()
    except TypeError as e:
        e_ = TypeError("set in memoization values must have sortable values")
        raise e_ from e

    dst.append("{")

    for s, _ in skey_keys:
        dst.append(s)
        dst.append(",")

    dst.append(")")


def stringify_atom(
    atom: object,
    dst: List[str],
    visited_container: Set[int],
    default: Callable[[object], object],
):
    """
    atom must not be a container (list/tuple/dict/set)
    """
    if atom is None:
        dst.append("None")
    elif isinstance(atom, str):
        dst.append(repr(atom))
    elif isinstance(atom, Complex):
        dst.append(repr(atom))
    elif isinstance(atom, bool):
        dst.append(repr(atom))
    elif isinstance(atom, bytes):
        dst.append(repr(atom))
    elif isinstance(atom, bytearray):
        dst.append(repr(atom))
    elif isinstance(atom, os.PathLike):
        dst.append(f"PathLike({repr(os.fspath(atom))})")
    else:
        try:
            v = default(atom)
        except TypeError:
            raise TypeError(f"{atom} is not serializable")

        _stringify(v, dst, visited_container, default)



from typing import (
    Any,
    Callable,
    Iterable,
    Mapping,
    Set,
    Type,
    Sequence,
)


T_Seq_Factory = Mapping[
    Type[Sequence[Any]], Callable[[Iterable[Any]], Sequence[Any]]
]
T_Map_Factory = Mapping[
    Type[Mapping[Any, Any]], Callable[[Mapping[Any, Any]], Mapping[Any, Any]]
]
T_Set_Factory = Mapping[Type[Set[Any]], Callable[[Iterable[Any]], Set[Any]]]


def map_structure(
    map_fn: Callable[[object], object],
    nest: Any,
    seq_factory: T_Seq_Factory = {list: list, tuple: tuple},
    map_factory: T_Map_Factory = {dict: dict},
    set_factory: T_Set_Factory = {set: set},
):
    assert callable(map_fn)

    def rec(nest: Any):
        for src, dst in seq_factory.items():
            if isinstance(nest, src):
                return dst(map(rec, nest))

        for src, dst in map_factory.items():
            if isinstance(nest, src):
                return dst({k: rec(v) for k, v in nest.items()})

        for src, dst in set_factory.items():
            if isinstance(nest, src):
                return dst(map(rec, nest))

        return map_fn(nest)

    return rec(nest)


def ordered_map_structure(
    map_fn: Callable[[object], object],
    nest: object,
    seq_factory: T_Seq_Factory = {list: list, tuple: tuple},
    map_factory: T_Map_Factory = {dict: dict},
    set_factory: T_Set_Factory = {set: set},
):
    def rec(nest: object):
        for src, dst in seq_factory.items():
            if isinstance(nest, src):
                return dst(map(rec, nest))

        for src, dst in map_factory.items():
            if isinstance(nest, src):
                keys = sorted(nest.keys())
                return dst({k: rec(nest[k]) for k in keys})

        for src, dst in set_factory.items():
            if isinstance(nest, src):
                values = sorted(nest)
                return dst(map(rec, values))

        return map_fn(nest)

    return rec(nest)



from .group_tree.groups import (
    UntypedGroup,
    StaticGroupBase,
    GroupsGroup,
    RulesGroup,
)
from .group_tree.core import make
from .group_tree.atom import Atom, Mem, Memstr, Memnone
from .group_tree.rule import Rule, SELF
from .group_tree.file import File, VFile, IFile
from .group_tree.graphviz import print_graphviz
from .group_tree.misc import print_method
from .core.make import MakeSummary

VERSION = "0.4.2"

__all__ = [
    "SELF",
    "UntypedGroup",
    "StaticGroupBase",
    "RulesGroup",
    "GroupsGroup",
    "make",
    "Atom",
    "Mem",
    "Memstr",
    "Memnone",
    "File",
    "VFile",
    "IFile",
    "print_graphviz",
    "print_method",
    "MakeSummary",
    "Rule",
]



from __future__ import annotations
from abc import ABCMeta, abstractmethod
import os
from pathlib import Path
from typing import (
    Callable,
    Generic,
    Iterable,
    Optional,
    Sequence,
    Set,
    TypeVar,
    Collection,
)

from .core.abc import IRule, UpdateResults, UpdateResult


class IMemo(metaclass=ABCMeta):
    @abstractmethod
    def compare(self, other: IMemo) -> bool:
        ...

    @abstractmethod
    def dumps(self) -> Iterable[bytes]:
        ...

    @classmethod
    @abstractmethod
    def loads(cls, data: bytes) -> IMemo:
        ...


TId = TypeVar("TId")

_T_method = TypeVar("_T_method", bound=Callable[[], object])


class Rule(IRule, Generic[TId, _T_method]):
    __slots__ = [
        "yfiles",
        "xfiles",
        "xfile_is_orig",
        "xfile_is_vf",
        "_deplist",
        "_method",
        "_id",
        "memo",
    ]

    def __init__(
        self,
        yfiles: Sequence[Path],
        xfiles: Sequence[Path],
        xfile_is_orig: Sequence[bool],
        xfile_is_vf: Sequence[bool],
        deplist: Set[int],
        method: _T_method,
        memo: IMemo,
        id: TId,
    ):
        assert len(xfiles) == len(xfile_is_orig)

        self.yfiles = yfiles
        self.xfiles = xfiles
        self.xfile_is_orig = xfile_is_orig
        self.xfile_is_vf = xfile_is_vf
        self._deplist = deplist
        self._method = method
        self._id = id
        self.memo = memo

    def check_update(self, par_updated: bool, dry_run: bool) -> UpdateResult:
        """
        Prerequisite: the y-list has at least one item.

        Procedure:

        - dry run:
            1. dry_run?
                yes: Any original x does not exist or has mtime of 0: Infeasible
                no:  Any x does not exist or has mtime of 0: Infeasible
            2. Any y is missing or has a mtime of 0: Necessary
            3. dry_run and any parent was updated: PossiblyNecessary
            4. Any x of type File is newer than the oldest y: Necessary
            5. Memoized values are updated: Necessary
            6. Otherwise: UpToDate
        """
        funcs = [
            _check_update_1,
            _check_update_2,
            _check_update_3,
            _check_update_4,
            _check_update_5,
        ]

        for func in funcs:
            res = func(
                ys=self.yfiles,
                xs=self.xfiles,
                xisorig=self.xfile_is_orig,
                xisvf=self.xfile_is_vf,
                dry_run=dry_run,
                par_updated=par_updated,
                memo=self.memo,
                old_memo_file=self.metadata_fname,
            )
            if res is not None:
                return res

        return UpdateResults.UpToDate()

    def preprocess(self):
        for f in self.yfiles:
            try:
                os.makedirs(f.parent, exist_ok=True)
            except Exception:
                pass

    def postprocess(self, succ: bool):
        if succ:
            self.update_memo()
        else:
            # set mtime to 0
            for f in self.yfiles:
                try:
                    os.utime(f, (0, 0))
                except Exception:
                    pass

            # delete vfile cache
            try:
                os.remove(self.metadata_fname)
            except Exception:
                pass

    @property
    def metadata_fname(self) -> Path:
        p = self.yfiles[0]
        return p.parent / ".jtcmake" / p.name

    def update_memo(self):
        os.makedirs(os.path.dirname(self.metadata_fname), exist_ok=True)
        with self.metadata_fname.open("wb") as f:
            f.writelines(self.memo.dumps())

    @property
    def method(self) -> _T_method:
        return self._method

    @property
    def deps(self):
        return self._deplist

    @property
    def id(self) -> TId:
        return self._id


def _check_update_1(
    xs: Sequence[Path],
    xisorig: Sequence[bool],
    dry_run: bool,
    **_: object,
) -> Optional[UpdateResult]:
    for f, isorig in zip(xs, xisorig):
        if not f.exists():
            if not dry_run or isorig:
                return UpdateResults.Infeasible(f"Input file {f} is missing")
        elif os.path.getmtime(f) == 0:
            if not dry_run or isorig:
                return UpdateResults.Infeasible(
                    f"Input file {f} has mtime of 0. Input files"
                    " with mtime of 0 are considered to be invalid."
                )

    return None


def _check_update_2(
    ys: Collection[Path], **_: object
) -> Optional[UpdateResult]:
    if any(not f.exists() for f in ys):
        return UpdateResults.Necessary()

    if any(os.path.getmtime(f) == 0 for f in ys):
        return UpdateResults.Necessary()


def _check_update_3(
    dry_run: bool, par_updated: bool, **_: object
) -> Optional[UpdateResult]:
    if dry_run and par_updated:
        return UpdateResults.PossiblyNecessary()


def _check_update_4(
    ys: Collection[Path], xs: Sequence[Path], xisvf: Sequence[bool], **_: object
) -> Optional[UpdateResult]:
    assert all(y.exists() for y in ys)
    getmtime = os.path.getmtime
    oldest_y = min(os.path.getmtime(f) for f in ys)
    if any(not isvf and getmtime(x) > oldest_y for x, isvf in zip(xs, xisvf)):
        return UpdateResults.Necessary()


def _check_update_5(
    memo: IMemo, old_memo_file: Path, **_: object
) -> Optional[UpdateResult]:
    if not os.path.exists(old_memo_file):
        return UpdateResults.Necessary()

    with old_memo_file.open("rb") as f:
        old_memo = memo.loads(f.read())

    if not memo.compare(old_memo):
        return UpdateResults.Necessary()



from __future__ import annotations
from abc import ABCMeta, abstractmethod
from typing import Any, Tuple, List

from ..utils.nest import map_structure


class IMemoAtom(metaclass=ABCMeta):
    @property
    @abstractmethod
    def memo_value(self) -> object:
        """
        Returns:
            object to be memoized
        """
        ...


class ILazyMemoValue(metaclass=ABCMeta):
    @abstractmethod
    def __call__(self) -> object:
        ...


class _EvalLazyMemoValues:
    __slots__ = ("lazy_memo_values",)

    def __init__(self, lazy_memo_values: List[ILazyMemoValue]):
        self.lazy_memo_values = lazy_memo_values

    def __call__(self) -> List[object]:
        return [v() for v in self.lazy_memo_values]


def unwrap_memo_values(nest: object) -> Tuple[object, _EvalLazyMemoValues]:
    lazy_values: List[ILazyMemoValue] = []

    def _unwrap_atom(atom: object):
        if isinstance(atom, IMemoAtom):
            v = atom.memo_value
            if isinstance(v, ILazyMemoValue):
                lazy_values.append(v)
                return None
            else:
                return v
        else:
            return atom

    nest = map_structure(_unwrap_atom, nest)

    return nest, _EvalLazyMemoValues(lazy_values)


class IAtom(IMemoAtom, metaclass=ABCMeta):
    @property
    @abstractmethod
    def real_value(self) -> object:
        ...


def unwrap_real_values(args: object) -> object:
    def repl(v: object) -> object:
        if isinstance(v, IAtom):
            return v.real_value
        else:
            return v

    return map_structure(repl, args)


class Atom(IAtom):
    __slots__ = ("_value", "_memo_value")

    def __init__(self, value: object, memo_value: object):
        """Create Atom: special object that can be included in args/kwargs
        of Group.add. Atom is used to:

        1. explicitly indicate an object being atom.
        2. specify memoized value for the object

        Args:
            value: object to be wrapped.
            memo_value: value used for memoization instead of `value`.

        Example1:

            Wrap a lambda function.

            ```
            g = create_group('root')
            g.add('rule.txt', func, Atom(lambda x: x**2, None))
            g.make()
            ```

            `func(Path("root/rule.txt"), lambda x: x ** 2)` will be executed.
            The lambda function will not be memoized (instead, `None` will be)


        Example2:

            Wrap a numpy array.

            ```
            array = np.array([1,2,3])
            g = create_group('root')
            g.add('rule.txt', method, Atom(array, str(array))
            g.make()
            ```

            `func(Path("root/rule.txt"), np.array([1,2,3])` will be executed.
            Instead of the ndarray object, "[1, 2, 3]" will be memoized.
        """
        self._value = value
        self._memo_value = memo_value

    @property
    def real_value(self):
        return self._value

    @property
    def memo_value(self):
        return self._memo_value

    def __repr__(self):
        v, m = repr(self.real_value), repr(self.memo_value)
        return f"Atom(value={v}, memo_value={m})"


def Mem(arg: object, memo_value: object) -> Any:
    return Atom(arg, memo_value)


def Memstr(arg: object) -> Any:
    """
    Alias for `Atom(arg, str(arg))`.
    Use str(arg) as the value for memoization of arg
    """
    return Atom(arg, str(arg))


def Memnone(arg: object) -> Any:
    """
    Alias for `Atom(arg, "")`.
    Let arg be not memoized.
    """
    return Atom(arg, "")



from __future__ import annotations
from os import PathLike
from abc import ABCMeta, abstractmethod
from typing import (
    Dict,
    Optional,
    Sequence,
    TypeVar,
    Union,
    Tuple,
    overload,
    Mapping,
    Callable,
)

from typing_extensions import ParamSpec

from ..file import File, VFile
from ..core import IGroup, IFile
from ..rule import parse_args_output_files, Rule_init_parse_deco_func, Rule
from ...utils.strpath import StrOrPath

P = ParamSpec("P")
K = TypeVar("K", bound=str)
_T_deco_f = TypeVar("_T_deco_f", bound=Callable[[], object])


class DynamicRuleContainerMixin(IGroup, metaclass=ABCMeta):
    @overload
    def add(
        self,
        name: StrOrPath,
        output_files: Union[
            Mapping[K, StrOrPath],
            Sequence[Union[K, PathLike[str]]],
            K,
            PathLike[str],
        ],
        method: Callable[P, object],
        /,
    ) -> Callable[P, Rule[str]]:
        ...

    @overload
    def add(
        self, name: StrOrPath, method: Callable[P, object], /
    ) -> Callable[P, Rule[str]]:
        ...

    def add(
        self, name: object, outs: object, method: object = None
    ) -> Callable[..., Rule[str]]:
        return self._add(name, outs, method, IFile_fact=File)

    @overload
    def addvf(
        self,
        name: StrOrPath,
        output_files: Union[
            Mapping[K, StrOrPath],
            Sequence[Union[K, PathLike[str]]],
            K,
            PathLike[str],
        ],
        method: Callable[P, object],
        /,
    ) -> Callable[P, Rule[str]]:
        ...

    @overload
    def addvf(
        self, name: StrOrPath, method: Callable[P, object], /
    ) -> Callable[P, Rule[str]]:
        ...

    def addvf(
        self, name: object, outs: object, method: object = None
    ) -> Callable[..., Rule[str]]:
        return self._add(name, outs, method, IFile_fact=VFile)

    def _add(
        self,
        name: object,
        outs: object,
        method: object = None,
        *,
        IFile_fact: Callable[[StrOrPath], IFile],
    ) -> Callable[..., Rule[str]]:
        if method is None:
            outs, method = name, outs

        if not isinstance(name, str):
            raise TypeError("name must be str or os.PathLike")

        name_ = str(name)

        outs_: Dict[str, IFile] = parse_args_output_files(
            name_, None, outs, IFile_fact
        )

        def _add(*args: object, **kwargs: object) -> Rule[str]:
            return self._add_rule(name_, outs_, method, args, kwargs)

        return _add

    def add_deco(
        self,
        name: StrOrPath,
        output_files: Optional[
            Union[
                Mapping[K, StrOrPath],
                Sequence[Union[str, PathLike[str]]],
                str,
                PathLike[str],
            ]
        ] = None,
        /,
    ) -> Callable[[_T_deco_f], _T_deco_f]:
        return self._add_deco(name, output_files, IFile_fact=File)

    def addvf_deco(
        self,
        name: StrOrPath,
        output_files: Optional[
            Union[
                Mapping[K, StrOrPath],
                Sequence[Union[str, PathLike[str]]],
                str,
                PathLike[str],
            ]
        ] = None,
        /,
    ) -> Callable[[_T_deco_f], _T_deco_f]:
        return self._add_deco(name, output_files, IFile_fact=VFile)

    def _add_deco(
        self,
        name: StrOrPath,
        output_files: Optional[
            Union[
                Mapping[K, StrOrPath],
                Sequence[Union[str, PathLike[str]]],
                str,
                PathLike[str],
            ]
        ] = None,
        *,
        IFile_fact: Callable[[StrOrPath], IFile],
    ) -> Callable[[_T_deco_f], _T_deco_f]:
        if output_files is None:
            output_files = name

        if not isinstance(name, str):
            raise TypeError("name must be str or os.PathLike")

        name_ = str(name)

        output_files_: Dict[str, IFile] = parse_args_output_files(
            name_, None, output_files, IFile_fact
        )

        def decorator(method: _T_deco_f):
            args, kwargs = Rule_init_parse_deco_func(method)
            self._add_rule(name_, output_files_, method, args, kwargs)
            return method

        return decorator

    def _add_rule(
        self,
        name: str,
        yfiles: Mapping[str, IFile],
        method: object,
        args: Tuple[object, ...],
        kwargs: Dict[str, object],
    ) -> Rule[str]:
        if name in self.rules:
            raise KeyError(
                f"A child rule with the same {name} already exists. "
                "All child groups and rules must have unique names"
            )

        if name in self.groups:
            raise KeyError(
                f"A child group with the same {name} already exists. "
                "All child groups and rules must have unique names"
            )

        def _factory() -> Rule[str]:
            r: Rule[str] = Rule.__new__(Rule)

            r.__init_at_once__(
                (*self.name_tuple, name),
                self._get_info(),
                self,
                yfiles,
                method,
                args,
                kwargs,
            )

            return r

        return self._add_rule_lazy(name, _factory)

    @abstractmethod
    def _add_rule_lazy(
        self, name: str, rule_factory: Callable[[], Rule[str]]
    ) -> Rule[str]:
        ...



from typing import Iterator, TypeVar, Mapping

K = TypeVar("K")
V = TypeVar("V")


class DictView(Mapping[K, V]):
    def __init__(self, dic: Mapping[K, V]):
        self._dic = dic

    def __getitem__(self, key: K) -> V:
        return self._dic[key]

    def __iter__(self) -> Iterator[K]:
        return self._dic.__iter__()

    def __len__(self) -> int:
        return len(self._dic)

    def __contains__(self, key: object) -> bool:
        return key in self._dic

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{dict(self)}"



from __future__ import annotations
import os
import sys
import base64
import hashlib
import pathlib
from typing import Dict, Tuple

from ..utils.strpath import StrOrPath
from .core import IFile
from .atom import IMemoAtom, ILazyMemoValue


if sys.platform == "win32":
    _Path = pathlib.WindowsPath
else:
    _Path = pathlib.PosixPath


class File(_Path, IFile):
    """
    An instance of this class represents a file, which can be an input or
    output of rules.

    When used as an input, on judging whether the rule must be updated,
    its modification time is compared to the modification time of the
    output files.
    If any of the output files is older than the input, the rule must be
    updated.
    """

    def is_value_file(self) -> bool:
        return False

    @property
    def memo_value(self) -> object:
        """
        Returns None.
        But the return value could be Path(self). Which is better?
        """
        return None


class _ContentHash(ILazyMemoValue):
    __slots__ = ["path"]

    def __init__(self, path: StrOrPath):
        self.path = path

    def __call__(self) -> str:
        return get_hash(self.path)


class VFile(_Path, IFile, IMemoAtom):
    """
    An instance of this class represents a value file, which can be
    an input or output of rules.

    When used as an input, on judging whether the rule must be updated,
    its modification time is compared to the modification time of the
    output files.
    If any of the output files is older than the input, the rule must be
    updated.
    """

    def is_value_file(self) -> bool:
        return True

    @property
    def memo_value(self) -> object:
        return _ContentHash(self)


_hash_cache: Dict[str, Tuple[float, str]] = {}


def get_hash(fname: StrOrPath) -> str:
    fname = os.path.realpath(fname)

    if fname in _hash_cache:
        if os.path.getmtime(fname) == _hash_cache[fname][0]:
            return _hash_cache[fname][1]

    mtime = os.path.getmtime(fname)

    with open(fname, "rb") as f:
        res = base64.b64encode(hashlib.md5(f.read()).digest()).decode()

    _hash_cache[os.path.abspath(fname)] = (mtime, res)

    return res



from __future__ import annotations
import inspect
from typing import (
    Any,
    Dict,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Generic,
    Union,
    get_origin,
    get_type_hints,
    Callable,
    overload,
)

from ..utils.strpath import StrOrPath
from ..utils.dict_view import DictView
from .core import IGroup, GroupTreeInfo, IRule
from .rule import Rule
from .group_mixins.basic import (
    BasicMixin,
    BasicInitMixin,
)
from .group_mixins.dynamic_container import DynamicRuleContainerMixin
from .group_mixins.memo import MemoMixin
from .group_mixins.selector import SelectorMixin


V = TypeVar("V")


class StaticGroupBase(BasicMixin, BasicInitMixin, SelectorMixin, MemoMixin):
    """
    Base class for static groups.

    A static group should be defined by subclassing StaticGroupBase.
    It must have type annotations to represent its child groups and rules.  ::

        class CustomStaticGroup(StaticGroupBase):
            child_rule1: Rule[str]
            child_rule2: Rule[Literal["a"]]
            child_group1: AnotherStaticGroup
            child_group2: RulesGroup

            '''
            Generic type parameters like ``[str]`` in this example is ignored
            at runtime. They are just hints for the type checker and IDE.
            '''

    The child nodes are automatically instanciated when the parent is
    instanciated. So you can read them without assigning values::
        
        g = CustomStaticGroup()
        print(g.child_rule1)
        print(g.child_group1)

    Remember that the child rules are automatically instanciated but not
    *initialized* . you have to manually initialize them with
    ``Rule.init`` ::
        
        g.child_rule1.init("child_file1.txt", copy)(souce_file, SELF)
    
    As a design pattern, it is recommended to have an initializer method
    that initializes the child rules and calls the initializer of the child
    groups to recursively initialize all the rules in the sub-tree. ::

        class CustomStaticGroup(StaticGroupBase):
            rule: Rule[str]
            group: AnotherStaticGroup
            
            def init(self, some_file: Path) -> CustomStaticGroup:
                # Initializer for this class. The method name "init" is not 
                # reserved so you can choose your own one.

                # Initialize the direct child rules
                self.rule.init(...)(...)

                # Initialize the child group (assuming ``AnotherStaticGroup``
                # has ``init`` to initialize itself)
                self.group.init(...)

                return self

        g = CustomStaticGroup().init(some_file)
        g.make()

    .. note::
       When you override the ``__init__`` method, you have to call
       ``super().__init__`` in it with appropriate arguments.

    """
    _info: GroupTreeInfo
    _name: Tuple[str, ...]
    _groups: Dict[str, IGroup]
    _rules: Dict[str, Rule[str]]
    _parent: IGroup

    def __init_as_child__(
        self,
        info: GroupTreeInfo,
        parent: IGroup,
        name: Tuple[str, ...],
    ):
        self._parent = parent
        self._info = info
        self._name = name

        self._groups = {}
        self._rules = {}

        for child_name, type_hint in get_type_hints(type(self)).items():
            fqcname = (*self._name, child_name)

            tp = _get_type(type_hint)

            if tp is None:
                continue

            if tp == Rule:
                # Rule
                r: Rule[str] = Rule.__new__(Rule)
                r.__init_partial__(fqcname, self._info, None, self)
                setattr(self, child_name, r)
                self._rules[child_name] = r
            elif issubclass(tp, IGroup) and not inspect.isabstract(tp):
                # Group
                g = tp.__new__(tp)
                g.__init_as_child__(self._info, self, fqcname)
                setattr(self, child_name, g)
                self._groups[child_name] = g

    @property
    def parent(self) -> IGroup:
        return self._parent

    @property
    def groups(self) -> Mapping[str, IGroup]:
        return DictView(self._groups)

    @property
    def rules(self) -> Mapping[str, IRule]:
        return DictView(self._rules)

    @property
    def name_tuple(self) -> Tuple[str, ...]:
        return self._name

    def _get_info(self) -> GroupTreeInfo:
        return self._info


T_Child = TypeVar("T_Child", bound=IGroup)


class GroupsGroup(
    BasicMixin, BasicInitMixin, SelectorMixin, MemoMixin, Generic[T_Child]
):
    _name: Tuple[str, ...]
    _parent: IGroup
    _info: GroupTreeInfo
    _groups: Dict[str, T_Child]
    _child_group_type: Union[None, Type[T_Child]] = None

    def __init_as_child__(
        self,
        info: GroupTreeInfo,
        parent: IGroup,
        name: Tuple[str, ...],
    ):
        self._info = info
        self._parent = parent
        self._name = name
        self._groups = {}

    def set_default_child(
        self, default_child_group_type: Type[T_Child]
    ) -> GroupsGroup[T_Child]:
        tp = _parse_child_group_type(default_child_group_type)
        self._child_group_type = tp  # pyright: ignore
        return self

    def set_props(
        self,
        default_child_group_type: Optional[Type[T_Child]],
        dirname: Optional[StrOrPath] = None,
        prefix: Optional[StrOrPath] = None,
    ) -> GroupsGroup[T_Child]:
        if default_child_group_type is not None:
            self.set_default_child(default_child_group_type)

        if dirname is not None or prefix is not None:
            self.set_prefix(dirname, prefix=prefix)  # pyright: ignore

        return self

    def add_group(
        self, name: str, child_group_type: Optional[Type[T_Child]] = None
    ) -> T_Child:
        if not isinstance(
            name, str
        ):  # pyright: ignore [reportUnnecessaryIsInstance]
            raise TypeError("name must be str")

        if name in self._groups:
            raise KeyError(f"Child group {name} already exists")

        if child_group_type is None:
            if self._child_group_type is None:
                raise Exception(
                    "No child group type is available. "
                    "You must provide `child_group_type` or, in advance, "
                    "set the default child group type by "
                    "`GroupsGroup.set_default_child(some_group_type)` or "
                    "`GroupsGroup.set_props(some_group_type)`. "
                )

            tp: Type[T_Child] = self._child_group_type
        else:
            tp = _parse_child_group_type(child_group_type)  # pyright: ignore

        g = tp.__new__(tp)
        g.__init_as_child__(self._info, self, (*self._name, name))

        self._groups[name] = g

        if name.isidentifier() and name[0] != "_" and not hasattr(self, name):
            setattr(self, name, g)

        return g

    @property
    def parent(self) -> IGroup:
        return self._parent

    @property
    def groups(self) -> Mapping[str, T_Child]:
        return DictView(self._groups)

    @property
    def rules(self) -> Mapping[str, IRule]:
        return DictView({})

    def __getitem__(self, k: str) -> T_Child:
        return self._groups[k]

    def __getattr__(self, k: str) -> T_Child:
        return self[k]

    @property
    def name_tuple(self) -> Tuple[str, ...]:
        return self._name

    def _get_info(self) -> GroupTreeInfo:
        return self._info


class RulesGroup(
    DynamicRuleContainerMixin,
    BasicMixin,
    BasicInitMixin,
    SelectorMixin,
    MemoMixin,
):
    _name: Tuple[str, ...]
    _parent: IGroup
    _info: GroupTreeInfo
    _rules: Dict[str, Rule[str]]

    def __init_as_child__(
        self, info: GroupTreeInfo, parent: IGroup, name: Tuple[str, ...]
    ):
        self._info = info
        self._parent = parent
        self._name = name
        self._rules = {}

    def _add_rule_lazy(
        self, name: str, rule_factory: Callable[[], Rule[str]]
    ) -> Rule[str]:
        if name in self._rules:
            raise KeyError(f"Rule {name} already exists in the group")

        r = rule_factory()

        self._rules[name] = r

        if name.isidentifier() and not hasattr(self, name):
            setattr(self, name, r)

        return r

    @property
    def parent(self) -> IGroup:
        return self._parent

    @property
    def groups(self) -> Mapping[str, IGroup]:
        return DictView({})

    @property
    def rules(self) -> Mapping[str, IRule]:
        return DictView(self._rules)

    def __getitem__(self, k: str) -> Rule[str]:
        return self._rules[k]

    def __getattr__(self, k: str) -> Rule[str]:
        return self[k]

    @property
    def name_tuple(self) -> Tuple[str, ...]:
        return self._name

    def _get_info(self) -> GroupTreeInfo:
        return self._info


class UntypedGroup(
    BasicMixin,
    DynamicRuleContainerMixin,
    BasicInitMixin,
    SelectorMixin,
    MemoMixin,
):
    _name: Tuple[str, ...]
    _parent: IGroup
    _info: GroupTreeInfo
    _groups: Dict[str, IGroup]
    _rules: Dict[str, Rule[str]]

    def __init_as_child__(
        self, info: GroupTreeInfo, parent: IGroup, name: Tuple[str, ...]
    ):
        self._info = info
        self._parent = parent
        self._name = name
        self._groups = {}
        self._rules = {}

    def _add_rule_lazy(
        self, name: str, rule_factory: Callable[[], Rule[str]]
    ) -> Rule[str]:
        if name in self._rules:
            raise KeyError(f"Rule {name} already exists in the group")

        r = rule_factory()

        self._rules[name] = r

        if name.isidentifier() and not hasattr(self, name):
            setattr(self, name, r)

        return r

    @overload
    def add_group(self, name: str, child_group_type: Type[T_Child]) -> T_Child:
        ...

    @overload
    def add_group(self, name: str) -> UntypedGroup:
        ...

    def add_group(self, name: object, child_group_type: Any = None) -> IGroup:
        if child_group_type is None:
            tp = UntypedGroup
        else:
            tp = _parse_child_group_type(child_group_type)

        if not isinstance(
            name, str
        ):  # pyright: ignore [reportUnnecessaryIsInstance]
            raise TypeError("name must be str")

        if name in self._groups:
            raise KeyError(
                f"A child group with the same {name} already exists. "
                "All child groups and rules must have unique names"
            )

        if name in self._rules:
            raise KeyError(
                f"A child rule with the same {name} already exists. "
                "All child groups and rules must have unique names"
            )

        g = tp.__new__(tp)
        g.__init_as_child__(self._info, self, (*self._name, name))

        self._groups[name] = g

        if name.isidentifier() and name[0] != "_" and not hasattr(self, name):
            setattr(self, name, g)

        return g

    @property
    def parent(self) -> IGroup:
        return self._parent

    @property
    def groups(self) -> Mapping[str, IGroup]:
        return DictView(self._groups)

    @property
    def rules(self) -> Mapping[str, IRule]:
        return DictView(self._rules)

    def _get_info(self) -> GroupTreeInfo:
        return self._info

    @property
    def name_tuple(self) -> Tuple[str, ...]:
        return self._name

    def __getitem__(self, k: str) -> Any:
        if k in self.groups:
            return self.groups[k]

        else:
            return self.rules[k]

    def __getattr__(self, __name: str) -> Any:
        if __name in self.groups:
            return self.groups[__name]
        elif __name in self.rules:
            return self.rules[__name]
        else:
            raise KeyError(f"No child group or rule named {__name}")


def _get_type(type_hint: object) -> Union[None, Type[Any]]:
    """
    Get instance of `type` from type hint (fully resolved one).
    On failure return None
    """
    if isinstance(type_hint, type):
        return type_hint

    origin = get_origin(type_hint)

    if isinstance(origin, type):
        return origin
    else:
        return None


def _parse_child_group_type(child_group_type: object) -> Type[IGroup]:
    tp = _get_type(child_group_type)

    if tp is None:
        raise TypeError(f"{tp} is not a valid type")

    if not issubclass(tp, IGroup):
        raise TypeError("Child group type must be a subclass of IGroup")

    if inspect.isabstract(tp):
        raise TypeError("Child group type must not be abstract")

    return tp



from typing import TypeVar
from ..core import INode
from ..atom import Atom

T = TypeVar("T")


class MemoMixin(INode):
    def mem(self: INode, value: T, memoized_value: object) -> T:
        self._get_info().memo_store[id(value)] = Atom(value, memoized_value)
        return value

    def memstr(self: INode, value: T) -> T:
        return MemoMixin.mem(self, value, str(value))

    def memnone(self: INode, value: T) -> T:
        return MemoMixin.mem(self, value, None)



from abc import ABCMeta, abstractmethod
import inspect
from typing import Callable, List, Optional, Sequence, Mapping, Dict, Tuple
from pathlib import Path

from ..raw_rule import Rule
from ..core.abc import IEvent
from ..core import events
from ..logwriter import IWriter, RichStr


class INoArgFunc(metaclass=ABCMeta):
    def __call__(self) -> object:
        self.method(*self.args, **self.kwargs)

    @property
    @abstractmethod
    def method(self) -> Callable[..., object]:
        ...

    @property
    @abstractmethod
    def args(self) -> Tuple[object, ...]:
        ...

    @property
    @abstractmethod
    def kwargs(self) -> Dict[str, object]:
        ...


def get_rule_name(r: Rule[int, INoArgFunc], id2name: Callable[[int], str]):
    return id2name(r.id)


def log_make_event(
    w: IWriter, e: IEvent[Rule[int, INoArgFunc]], id2name: Callable[[int], str]
):
    if isinstance(e, events.ErrorRuleEvent):
        # Show stack trace and error message
        w.warning("".join(e.trace_exc.format()))

        r = e.rule

        name = RichStr(get_rule_name(r, id2name), c=(0, 0xCC, 0))

        if isinstance(e, events.PreProcError):
            w.error(
                "Failed to make ",
                name,
                ": An error occured during preprocessing",
            )
        elif isinstance(e, events.ExecError):
            w.error(
                "Failed to make ",
                name,
                f": Method {get_func_name(r.method)} failed",
            )
        elif isinstance(e, events.PostProcError):
            w.error(
                f"Failed to make {name}: "
                f"An error occured during post-processing. "
                f"Make sure to remove the output files (if any) "
                f"by yourself"
            )
        elif isinstance(e, events.FatalError):
            w.error("Fatal error")
        else:
            w.warning(f"Unhandled error event for {r}")

        return
    elif isinstance(e, events.RuleEvent):
        r = e.rule

        name = RichStr(get_rule_name(r, id2name), c=(0, 0xCC, 0))

        if isinstance(e, events.Skip):
            if e.is_direct_target:
                w.info("Skip ", name)
            else:
                w.debug("Skip ", name)
        elif isinstance(e, events.Start):
            msg = []
            method_ = r.method
            tostrs_func_call(msg, method_.method, method_.args, method_.kwargs)
            msg = add_indent(msg, "  ")
            w.info("Make ", name, "\n", *msg)
        elif isinstance(e, events.Done):
            w.info("Done ", name)
        elif isinstance(e, events.DryRun):
            msg = []
            method_ = r.method
            tostrs_func_call(msg, method_.method, method_.args, method_.kwargs)
            msg = add_indent(msg, "  ")
            w.info("Make (dry) ", name, "\n", *msg)
        elif isinstance(e, events.UpdateInfeasible):
            w.error(
                "Cannot make ",
                name,
                f": {e.reason}",
            )
        else:
            w.warning(f"Unhandled event for {r}")
        return
    elif isinstance(e, events.StopOnFail):
        w.warning("Execution aborted due to an error")
    else:
        w.warning(f"Unhandled event {e}")


def add_indent(sl: Sequence[str], indent: str) -> List[str]:
    res: List[str] = []
    for i, s in enumerate(sl):
        if i == 0 or (i > 0 and len(sl[i - 1]) > 0 and sl[i - 1][-1] == "\n"):
            res.append(indent + s)
        else:
            res.append(s)
    return res


def get_func_name(f: Callable[..., object]) -> str:
    try:
        name, mod = f.__qualname__, f.__module__

        if mod == "builtins" or mod == "__main__":
            return name
        else:
            return f"{mod}.{name}"
    except Exception:
        return "<unkonw function>"


def tostrs_func_call(
    dst: List[str],
    f: Callable[..., object],
    args: Tuple[object, ...],
    kwargs: Dict[str, object],
):
    bn = inspect.signature(f).bind(*args, **kwargs)
    bn.apply_defaults()

    dst.append(RichStr(get_func_name(f), c=(0, 0x80, 0xFF)))
    dst.append("(\n")

    for k, v in bn.arguments.items():
        dst.append(RichStr(f"  {k}", c=(0xFF, 0x80, 0)))
        dst.append(" = ")
        tostrs_obj(dst, v, capacity=500)
        dst.append(",\n")
    dst.append(")\n")


def tostrs_obj(dst: List[str], o: object, capacity: Optional[int] = None):
    _tostrs_obj(dst, o, capacity or 10**10)


def _tostrs_obj(dst: List[str], o: object, capacity: int) -> int:
    if isinstance(o, (tuple, list)):
        dst.append("[")
        for i, v in enumerate(o):  # pyright: ignore [reportUnknownVariableType]
            if capacity <= 0:
                dst.append(", ...")
                capacity -= 5
                break
            capacity = _tostrs_obj(dst, v, capacity)
            if i < len(o) - 1:
                dst.append(", ")
                capacity -= 2
        dst.append("]")
        return capacity
    elif isinstance(o, Mapping):
        dst.append("{")
        for i, (
            k,  # pyright: ignore [reportUnknownVariableType]
            v,  # pyright: ignore [reportUnknownVariableType]
        ) in enumerate(o.items()):
            if capacity <= 0:
                dst.append(", ... ")
                capacity -= 5
                break
            dst.append(repr(k) + ": ")
            capacity = _tostrs_obj(dst, v, capacity)
            if i < len(o) - 1:
                dst.append(", ")
                capacity -= 2
        dst.append("}")
        return capacity
    elif isinstance(o, Path):
        res = RichStr(repr(o), link=str(o))
        dst.append(res)
        return capacity - len(res)
    else:
        res = repr(o)
        if len(res) > capacity:
            res = res[: capacity // 2] + " ... " + res[-capacity // 2 :]
        dst.append(res)
        return capacity - len(res)



from __future__ import annotations
from abc import ABCMeta, abstractmethod
import os
import itertools
from typing import (
    Callable,
    Collection,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Dict,
    Union,
    overload,
    Sequence,
    List,
    Set,
    final,
)
from typing_extensions import ParamSpec, Concatenate
from pathlib import Path

from ..raw_rule import Rule as _RawRule, IMemo
from ..core.make import MakeSummary, make as _make
from ..core.make_mp import make_mp_spawn
from ..core.abc import IEvent
from ..logwriter import IWriter
from .event_logger import log_make_event, INoArgFunc
from ..utils.strpath import StrOrPath, fspath2str
from .atom import IAtom


class INode(metaclass=ABCMeta):
    @property
    @abstractmethod
    def parent(self) -> IGroup:
        ...

    @abstractmethod
    def _get_info(self) -> GroupTreeInfo:
        ...

    @property
    @abstractmethod
    def name_tuple(self) -> Tuple[str, ...]:
        ...

    @property
    def name(self) -> str:
        if len(self.name_tuple) == 0:
            return "/"
        else:
            return "/" + "/".join(self.name_tuple)


T_Self = TypeVar("T_Self", bound="IGroup")


class IGroup(INode, metaclass=ABCMeta):
    __prefix: Union[None, str] = None

    @abstractmethod
    def __init_as_child__(
        self,
        info: GroupTreeInfo,
        parent: IGroup,
        name: Tuple[str, ...],
    ):
        ...

    @overload
    def set_prefix(
        self: T_Self, dirname: StrOrPath, *, prefix: None = None
    ) -> T_Self:
        ...

    @overload
    def set_prefix(
        self: T_Self, dirname: None = None, *, prefix: StrOrPath
    ) -> T_Self:
        ...

    @final
    def set_prefix(
        self: T_Self, dirname: object = None, *, prefix: object = None
    ) -> T_Self:
        if self.__prefix is not None:
            raise Exception(
                f'Prefix is already set (to "{self.__prefix}") and '
                "may not be overwritten. Make sure to call set_prefix "
                "before initializing child groups and rules. "
            )

        p = parse_args_prefix(dirname, prefix)

        if self.parent == self:
            self.__prefix = p
        else:
            self.__prefix = concat_prefix(p, self.parent.prefix)

        return self

    @final
    @property
    def prefix(self) -> str:
        if self.__prefix is None:
            # Root node must get prefix in __init__
            assert self.parent != self

            self.set_prefix(self.name_tuple[-1])
            assert self.__prefix is not None
            return self.__prefix
        else:
            return self.__prefix

    @property
    @abstractmethod
    def groups(self) -> Mapping[str, IGroup]:
        ...

    @property
    @abstractmethod
    def rules(self) -> Mapping[str, IRule]:
        ...


class IRule(INode, metaclass=ABCMeta):
    @property
    @abstractmethod
    def raw_rule_id(self) -> int:
        ...

    @property
    @abstractmethod
    def files(self) -> Mapping[str, IFile]:
        ...

    @property
    @abstractmethod
    def xfiles(self) -> Collection[str]:
        ...

    @abstractmethod
    def clean(self) -> None:
        ...

    @abstractmethod
    def touch(
        self, file: bool, memo: bool, create: bool, t: Union[float, None]
    ) -> None:
        ...


class IFile(Path, IAtom, metaclass=ABCMeta):
    """
    Abstract base class to represent a file object.
    """

    """
    For implementors of this ABC:
        It is highly recommended not to have variable properties (public or
        private) in the new class because the default implementations of the
        generative methods of Path (absolute(), resolve(), etc) create new
        instance without copying subclasses' variable properties.
    """

    @abstractmethod
    def is_value_file(self) -> bool:
        ...

    def __eq__(self, other: object) -> bool:
        ts, to = type(self), type(other)
        if issubclass(to, ts) or issubclass(ts, to):
            return super().__eq__(other)
        else:
            return False

    @property
    def real_value(self) -> object:
        return Path(self)


class RuleStore:
    __slots__ = (
        "rules",
        "ypath2idx",
        "idx2xpaths",
        "path2file",
        "idx2name",
    )

    rules: List[_RawRule[int, INoArgFunc]]
    ypath2idx: Dict[str, int]
    idx2xpaths: Dict[int, Sequence[str]]
    path2file: Dict[str, IFile]
    idx2name: Dict[int, Tuple[str, ...]]

    def __init__(self):
        self.rules = []
        self.ypath2idx = {}
        self.idx2xpaths = {}
        self.path2file = {}
        self.idx2name = {}

    def add(
        self,
        yp2f: Mapping[str, IFile],  # abspath(str) => IFile
        xp2f: Mapping[str, IFile],  # abspath(str) => IFile
        method: INoArgFunc,
        memo: IMemo,
        name: Tuple[str, ...],
    ) -> _RawRule[int, INoArgFunc]:
        # Check duplicated registration of yfiles
        for p, f in yp2f.items():
            if p in self.ypath2idx:
                raise ValueError(
                    f"File {f} is already used as an output of another rule"
                )

        # Check IFile type consistency of xfiles
        for p, f in xp2f.items():
            f_ = self.path2file.get(p)
            if f_ is not None and f_.resolve() != f.resolve():
                raise TypeError(
                    f"IFile inconsistency detected: argument {f} is of type "
                    f"{type(f)} but the file was registered to be created as "
                    f"{type(f_)}"
                )

        # Create Rule
        id = len(self.rules)
        xids = [self.ypath2idx.get(p, -1) for p in xp2f]
        rule = _RawRule(
            yfiles=list(yp2f.values()),
            xfiles=list(xp2f.values()),
            xfile_is_orig=[i == -1 for i in xids],
            xfile_is_vf=[f.is_value_file() for f in xp2f.values()],
            deplist=set(xids) - {-1},
            method=method,
            memo=memo,
            id=id,
        )

        # Update stores
        self.rules.append(rule)

        for p, f in yp2f.items():
            self.ypath2idx[p] = id
            self.path2file[p] = f

        for p, f in xp2f.items():
            if p not in self.ypath2idx:
                # Add original
                self.ypath2idx[p] = -1
                self.path2file[p] = f

        self.idx2xpaths[id] = list(xp2f)

        self.idx2name[id] = name

        return rule


class GroupTreeInfo:
    __slots__ = (
        "rule_store",
        "logwriter",
        "memo_factory",
        "memo_store",
        "rules_to_be_init",
    )

    rule_store: RuleStore
    logwriter: IWriter
    memo_factory: Callable[[object], IMemo]
    memo_store: Dict[int, IAtom]
    rules_to_be_init: Set[Tuple[str, ...]]

    def __init__(
        self, logwriter: IWriter, memo_factory: Callable[[object], IMemo]
    ):
        self.logwriter = logwriter
        self.memo_factory = memo_factory
        self.memo_store = {}
        self.rule_store = RuleStore()
        self.rules_to_be_init = set()


P = ParamSpec("P")
T = TypeVar("T")
T_INode = TypeVar("T_INode", bound=INode)


def require_tree_init(
    method: Callable[Concatenate[T_INode, P], T]
) -> Callable[Concatenate[T_INode, P], T]:
    def _method(self: T_INode, *args: P.args, **kwargs: P.kwargs) -> T:
        info = self._get_info()  # pyright: ignore [reportPrivateUsage]

        if len(info.rules_to_be_init) != 0:
            top10names = itertools.islice(info.rules_to_be_init, 10)
            top10names = ["/" + "/".join(n) for n in top10names]
            raise RuntimeError(
                "All rules in the group tree must be initialized "
                f"before calling this method. {len(info.rules_to_be_init)} "
                f"rules are not initialized. {len(top10names)} of them are: "
                "\n" + "\n".join(top10names)
            )
        return method(self, *args, **kwargs)

    return _method


def make(
    *rule_or_groups: Union[IGroup, IRule],
    dry_run: bool = False,
    keep_going: bool = False,
    njobs: Optional[int] = None,
) -> MakeSummary:
    """make rules

    Args:
        rules_or_groups (Sequence[RuleNodeBase|Group]):
            Rules and Groups containing target Rules
        dry_run:
            instead of actually excuting methods, print expected execution logs.
        keep_going:
            If False (default), stop everything when a rule fails.
            If True, when a rule fails, keep executing other rules
            except the ones depend on the failed rule.
        njobs:
            Maximum number of rules that can be made concurrently.
            Defaults to 1 (single process, single thread).

            Note that safely using njobs >= 2 and fully exploiting the power
            of multi-core processors require a certain level of
            understanding of Python's threading and multiprocessing.

            Each rule is made on a child process if it is transferable.
            A rule is "transferable" if both of the following conditions
            are met:

            - method/args/kwargs of the rule are all picklable
            - Pickle representation of method/args/kwargs created in the
              main process is unpicklable in child processes

            If a rule is not transferable, it is made on a sub-thread of
            the main process. Thus the method must be thread-safe. Also note
            that methods running on the main process are subject to the
            global interpreter lock (GIL) constraints.

            Child processes are started by the 'spawn' method, not 'fork',
            even on Linux systems.
            njobs >= 2 may not work on interactive interpreters.
            It should work on Jupyter Notebook/Lab but any function or class
            that are defined on the notebook is not transferable and thus
            executed in the main process.
    """
    if len(rule_or_groups) == 0:
        return MakeSummary(total=0, update=0, skip=0, fail=0, discard=0)

    for node in rule_or_groups:
        if not isinstance(
            node, (IGroup, IRule)
        ):  # pyright: ignore [reportUnnecessaryIsInstance]
            raise TypeError("Invalid node {node}")

    info = get_group_info_of_nodes(rule_or_groups)

    ids = gather_raw_rule_ids(rule_or_groups)

    def callback_(event: IEvent[_RawRule[int, INoArgFunc]]):
        def id2name(i: int) -> str:
            return "/".join(info.rule_store.idx2name[i])

        log_make_event(info.logwriter, event, id2name)

    if njobs is not None and njobs >= 2:
        return make_mp_spawn(
            info.rule_store.rules, ids, dry_run, keep_going, callback_, njobs
        )
    else:
        return _make(info.rule_store.rules, ids, dry_run, keep_going, callback_)


def get_group_info_of_nodes(nodes: Sequence[INode]) -> GroupTreeInfo:
    if len(nodes) == 0:
        raise Exception("Internal error: nodes must not be empty")

    info = nodes[0]._get_info()  # pyright: ignore [reportPrivateUsage]

    for node in nodes:
        _info = node._get_info()  # pyright: ignore [reportPrivateUsage]
        if _info is not info:
            raise ValueError(
                "All Groups/Rules must belong to the same Group tree. "
            )

    return info


def gather_raw_rule_ids(
    group_or_rules: Sequence[Union[IGroup, IRule]]
) -> List[int]:
    ids: List[int] = []
    visited: Set[INode] = set()

    stack = list(reversed(group_or_rules))

    while stack:
        node = stack.pop()
        if node in visited:
            continue

        visited.add(node)

        if isinstance(node, IRule):
            ids.append(node.raw_rule_id)
        else:
            stack.extend(node.groups.values())
            stack.extend(node.rules.values())

    return ids


def parse_args_prefix(dirname: object, prefix: object) -> str:
    if dirname is not None and prefix is not None:
        raise TypeError(
            "Either dirname or prefix, but not both must be specified"
        )

    if dirname is not None:
        if not isinstance(dirname, (str, os.PathLike)):
            raise TypeError("dirname must be str or PathLike")

        prefix_ = fspath2str(dirname) + os.path.sep
    else:
        if prefix is None:
            prefix_ = ""
        elif isinstance(prefix, (str, os.PathLike)):
            prefix_ = fspath2str(prefix)
        else:
            raise TypeError("prefix must be str or PathLike")

    return prefix_


def concat_prefix(base: str, prefix: str) -> str:
    base = os.path.expanduser(base)
    return base if os.path.isabs(base) else prefix + base
